# coding: utf-8

from supervisely_lib.geometry.geometry import Geometry
from supervisely_lib.geometry.constants import ANY_SHAPE


class AnyGeometry(Geometry):
    '''
    This is a class for creating and using AnyGeometry for Labels.
    '''
    @staticmethod
    def geometry_name():
        return ANY_SHAPE



# coding: utf-8
import uuid

from supervisely_lib.api.module_api import ApiField
from supervisely_lib.app_widget.widget_type import WidgetType
from supervisely_lib.app_widget.base_widget import BaseWidget


class ButtonWidget(BaseWidget):
    widget_type = WidgetType.BUTTON

    @classmethod
    def create(cls, name, description, command, id=None):
        res = dict()
        res[ApiField.TYPE] = str(cls.widget_type)
        res[ApiField.NAME] = "button block title"
        res[ApiField.SUBTITLE] = "button block description"
        res[ApiField.CONTENT] = {
            ApiField.TITLE: name,
            ApiField.DESCRIPTION: description,
            ApiField.COMMAND: command,
        }
        res[ApiField.ID] = uuid.uuid4().hex if id is None else id
        return res


# coding: utf-8

from supervisely_lib.geometry.rectangle import Rectangle
from supervisely_lib.geometry.validation import is_2d_int_coords_valid


class SlidingWindows:
    def __init__(self, window_shape, min_overlap):
        if not is_2d_int_coords_valid([window_shape]):
            raise ValueError('window_shape must contains 2 integers.')

        if not is_2d_int_coords_valid([min_overlap]):
            raise ValueError('min_overlap must contains 2 integers.')

        self.window_shape = tuple(window_shape)
        self.min_overlap = tuple(min_overlap)
        self.stride = tuple(self.window_shape[i] - self.min_overlap[i] for i in (0, 1))
        if min(self.stride) < 1:
            raise RuntimeError('Wrong sliding window settings, overlap is too high.')

    def get(self, source_shape):
        source_rect = Rectangle.from_size(source_shape)
        window_rect = Rectangle.from_size(self.window_shape)
        if not source_rect.contains(window_rect):
            raise RuntimeError('Sliding window: window is larger than source (image).')

        hw_limit = tuple(source_shape[i] - self.window_shape[i] for i in (0, 1))
        for wind_top in range(0, hw_limit[0] + self.stride[0], self.stride[0]):
            for wind_left in range(0, hw_limit[1] + self.stride[1], self.stride[1]):
                window_top = min(wind_top, hw_limit[0])
                window_left = min(wind_left, hw_limit[1])
                roi = window_rect.translate(drow=window_top, dcol=window_left)
                if not source_rect.contains(roi):
                    raise RuntimeError('Sliding window: result crop bounds are invalid.')
                yield roi



# coding: utf-8

import os
import shutil
import errno
import tarfile
import subprocess
import requests
from requests.structures import CaseInsensitiveDict

from supervisely_lib._utils import get_bytes_hash, get_string_hash
from supervisely_lib.io.fs_cache import FileCache


def get_file_name(path: str) -> str:
    """
    Extracts file name from a given path.

    Args:
        path: File path.
    Returns:
         File name.
    """
    return os.path.splitext(os.path.basename(path))[0]


def get_file_ext(path: str) -> str:
    """
    Extracts file extension from a given path.

    Args:
        path: File path.
    Returns:
         File extension.
    """
    return os.path.splitext(os.path.basename(path))[1]


def get_file_name_with_ext(path: str) -> str:
    """
    Extracts file name with ext from a given path.

    Args:
        path: File path.
    Returns:
         File name with ext.
    """
    return os.path.basename(path)


def list_dir_recursively(dir: str) -> list:
    """
    Recursively walks through directory and returns list with all file paths.

    Args:
        dir: Target dir path.
    Returns:
         A list containing file paths.
    """
    all_files = []
    for root, dirs, files in os.walk(dir):
        for name in files:
            file_path = os.path.join(root, name)
            file_path = os.path.relpath(file_path, dir)
            all_files.append(file_path)
    return all_files


def list_files_recursively(dir: str, valid_extensions: list = None, filter_fn=None) -> list:
    """
    Recursively walks through directory and returns list with all file paths.

    Args:
        dir: Target dir path.
        valid_extensions:
        filter_fn: function with a single argument that determines whether to keep a given file path.
    Returns:
         A list containing file paths.
    """

    def file_path_generator():
        for dir_name, _, file_names in os.walk(dir):
            for filename in file_names:
                yield os.path.join(dir_name, filename)

    return [file_path for file_path in file_path_generator() if
            (valid_extensions is None or get_file_ext(file_path) in valid_extensions) and
            (filter_fn is None or filter_fn(file_path))]


def list_files(dir: str, valid_extensions: list = None, filter_fn=None) -> list:
    """
    Returns list with file paths presented in given directory.

    Args:
        dir: Target dir path.
        valid_extensions:
        filter_fn: function with a single argument that determines whether to keep a given file path.
    Returns:
         A list containing file paths.
    """
    res = list(os.path.join(dir, x.name) for x in os.scandir(dir) if x.is_file())
    return [file_path for file_path in res if
      (valid_extensions is None or get_file_ext(file_path) in valid_extensions) and
      (filter_fn is None or filter_fn(file_path))]


def mkdir(dir: str):
    """
    Creates a leaf directory and all intermediate ones.

    Args:
        dir: Target directory path.

    """
    os.makedirs(dir, mode=0o777, exist_ok=True)


# create base path recursively; pay attention to slash-terminating paths
def ensure_base_path(path):
    dst_dir = os.path.split(path)[0]
    if dst_dir:
        mkdir(dst_dir)


def copy_file(src: str, dst: str):
    """
    Args:
        src: Source file path.
        dst: Destination file path.

    """
    ensure_base_path(dst)
    with open(dst, 'wb') as out_f:
        with open(src, 'rb') as in_f:
            shutil.copyfileobj(in_f, out_f, length=1024 * 1024)


def hardlink_or_copy_file(src: str, dst: str):
    try:
        os.link(src, dst)
    except OSError:
        copy_file(src, dst)


def hardlink_or_copy_tree(src: str, dst: str):
    mkdir(dst)
    for dir_name, _, file_names in os.walk(src):
        relative_dir = os.path.relpath(dir_name, src)
        dst_sub_dir = os.path.join(dst, relative_dir)
        mkdir(dst_sub_dir)
        for file_name in file_names:
            hardlink_or_copy_file(os.path.join(dir_name, file_name), os.path.join(dst_sub_dir, file_name))


def dir_exists(dir: str) -> bool:
    """
    Check whether directory exists or not.

    Args:
        dir: Target directory path.
    Returns:
        True if directory exists, False otherwise.
    """
    return os.path.isdir(dir)


def dir_empty(dir: str) -> bool:
    """
    Check whether directory is empty or not.

    Args:
        dir: Target directory path.
    Returns:
        True if directory is empty, False otherwise.
    """
    if dir_exists(dir) and len(list_files_recursively(dir)) > 0:
        return False
    return True


def file_exists(path: str) -> bool:
    """
    Check whether file exists or not.

    Args:
        path: File path.
    Returns:
        True if file exists, False otherwise.
    """
    return os.path.isfile(path)


def get_subdirs(dir_path: str) -> list:
    """
    Return a list containing the names of the directories in the given directory.

    Args:
        dir_path: Target directory path.
    Returns:
        List containing directories names.
    """
    res = list(x.name for x in os.scandir(dir_path) if x.is_dir())
    return res


# removes directory content recursively
def clean_dir(dir_: str):
    """
    Recursively delete a directory tree, but save root directory.
    Args:
        dir_: Target directory path.
    """
    shutil.rmtree(dir_, ignore_errors=True)
    mkdir(dir_)


def remove_dir(dir_: str):
    """
    Recursively delete a directory tree.
    Args:
        dir_: Target directory path.
    """
    shutil.rmtree(dir_, ignore_errors=True)


def silent_remove(file_path: str):
    """
    Remove file which may not exist.
    Args:
        file_path: File path.
    """
    try:
        os.remove(file_path)
    except OSError as e:
        if e.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
            raise


def get_file_size(path: str):
    """
    Return the size of a file, reported by os.stat().
    Args:
        path: File path.
    """
    return os.path.getsize(path)


def get_directory_size(dir_path: str) -> int:
    """
    Return the size of a directory.

    Args:
        dir_path: Target directory path.
    Returns:
         Directory size in bytes.
    """
    total_size = 0
    for dirpath, _, filenames in os.walk(dir_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += get_file_size(fp)
    return total_size


def archive_directory(dir_: str, tar_path: str):
    """
    Create tar archive from directory.
    Args:
        dir_: Target directory path.
        tar_path: Path for output tar archive.
    """
    with tarfile.open(tar_path, 'w', encoding='utf-8') as tar:
        tar.add(dir_, arcname=os.path.sep)


def get_file_hash(path):
    return get_bytes_hash(open(path, 'rb').read())


def tree(dir_path):
    out = subprocess.Popen(['tree', '--filelimit', '500', '-h', '-n', dir_path],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT)
    stdout, stderr = out.communicate()
    return stdout.decode("utf-8")


def log_tree(dir_path, logger):
    out = tree(dir_path)
    logger.info("DIRECTORY_TREE", extra={'tree': out})


def touch(path):
    ensure_base_path(path)
    with open(path, 'a'):
        os.utime(path, None)


def download(url, save_path, cache: FileCache = None, progress=None):
    def _download():
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            total_size_in_bytes = int(CaseInsensitiveDict(r.headers).get('Content-Length', '0'))
            if progress is not None:
                progress.set(0, total_size_in_bytes)
            with open(save_path, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
                    if progress is not None:
                        progress.iters_done_report(len(chunk))

    if cache is None:
        _download()
    else:
        cache_path = cache.check_storage_object(get_string_hash(url), get_file_ext(save_path))
        if cache_path is None:
            # file not in cache
            _download()
            cache.write_object(save_path, get_string_hash(url))
        else:
            cache.read_object(get_string_hash(url), save_path)
            if progress is not None:
                progress.set(0, get_file_size(save_path))
                progress.iters_done_report(get_file_size(save_path))

    return save_path


# coding: utf-8

import math

from supervisely_lib.sly_logger import logger, EventType
from supervisely_lib._utils import sizeof_fmt

# float progress of training, since zero
def epoch_float(epoch, train_it, train_its):
    return epoch + train_it / float(train_its)


class Progress:
    '''
    This is a class for conveniently monitoring the operation of modules and displaying statistics on data processing
    '''
    def __init__(self, message, total_cnt, ext_logger=None, is_size=False, need_info_log=False):
        '''
        :param message: str
        :param total_cnt: int
        :param ext_logger: Logger class object
        '''
        self.is_size = is_size
        self.message = message
        self.total = total_cnt
        self.current = 0
        self.is_total_unknown = total_cnt == 0

        self.total_label = ''
        self.current_label = ''
        self._refresh_labels()

        self.reported_cnt = 0
        self.logger = logger if ext_logger is None else ext_logger
        self.report_every = max(1, math.ceil(total_cnt / 100))
        self.need_info_log = need_info_log

        if self.is_size and self.is_total_unknown:
            self.report_every = 5 * 1024 * 1024  # 5mb

        self.report_progress()

    def _refresh_labels(self):
        if self.is_size:
            self.total_label = sizeof_fmt(self.total) if self.total > 0 else sizeof_fmt(self.current)
            self.current_label = sizeof_fmt(self.current)
        else:
            self.total_label = str(self.total if self.total > 0 else self.current)
            self.current_label = str(self.current)

    def iter_done(self):
        '''
        Increments the current iteration counter by 1
        '''
        self.current += 1
        if self.is_total_unknown:
            self.total = self.current
        self._refresh_labels()

    def iters_done(self, count):
        '''
        Increments the current iteration counter by given count
        :param count: int
        '''
        self.current += count
        if self.is_total_unknown:
            self.total = self.current
        self._refresh_labels()

    def report_progress(self):
        '''
        Logs a message with level INFO on logger. Message contain type of progress, subtask message, currtnt and total number of iterations
        '''
        extra = {
            'event_type': EventType.PROGRESS,
            'subtask': self.message,
            'current': math.ceil(self.current),
            'total': math.ceil(self.total) if self.total > 0 else math.ceil(self.current),
        }

        if self.is_size:
            extra['current_label'] = self.current_label
            extra['total_label'] = self.total_label

        self.logger.info('progress', extra=extra)
        self.reported_cnt += 1

        if self.need_info_log is True:
            self.logger.info(f"{self.message} [{self.current_label} / {self.total_label}]")

    def need_report(self):
        if (self.current == self.total) \
                or (self.current % self.report_every == 0) \
                or ((self.reported_cnt - 1) < (self.current // self.report_every)):
            return True
        return False

    def report_if_needed(self):
        '''
        The function determines whether the message should be logged depending on current number of iterations
        '''
        if self.need_report():
            self.report_progress()

    def iter_done_report(self):  # finish & report
        '''
        Increments the current iteration counter by 1 and logs a message depending on current number of iterations
        :return:
        '''
        self.iter_done()
        self.report_if_needed()

    def iters_done_report(self, count):  # finish & report
        '''
        Increments the current iteration counter by given count and logs a message depending on current number of iterations
        :param count: int
        '''
        self.iters_done(count)
        self.report_if_needed()

    #@TODO: to upload dtl archive
    def set_current_value(self, value):
        '''
        Increments the current iteration counter by this value minus the current value of the counter and logs a message depending on current number of iterations
        :param value: int
        '''
        self.iters_done_report(value - self.current)

    def set(self, current, total):
        self.total = total
        self.current = current
        self.reported_cnt = 0
        self.report_every = max(1, math.ceil(total / 100))
        self._refresh_labels()
        self.report_if_needed()


def report_agent_rpc_ready():
    '''
    Logs a message with level INFO on logger
    '''
    logger.info('Ready to get events', extra={ 'event_type': EventType.TASK_DEPLOYED })


def report_import_finished():
    '''
    Logs a message with level INFO on logger
    '''
    logger.info('import finished', extra={'event_type': EventType.IMPORT_APPLIED})


def report_inference_finished():
    '''
    Logs a message with level INFO on logger
    '''
    logger.info('model applied', extra={'event_type': EventType.MODEL_APPLIED})


def report_dtl_finished():
    '''
    Logs a message with level INFO on logger
    '''
    logger.info('DTL finished', extra={'event_type': EventType.DTL_APPLIED})


def report_dtl_verification_finished(output):
    '''
    Logs a message with level INFO on logger
    :param output: str
    '''
    logger.info('Verification finished.', extra={'output': output, 'event_type': EventType.TASK_VERIFIED})


def _report_metrics(m_type, epoch, metrics):
    logger.info('metrics', extra={
        'event_type': EventType.METRICS,
        'type': m_type,
        'epoch': epoch,
        'metrics': metrics
    })


def report_metrics_training(epoch, metrics):
    _report_metrics('train', epoch, metrics)


def report_metrics_validation(epoch, metrics):
    _report_metrics('val', epoch, metrics)


def report_checkpoint_saved(checkpoint_idx, subdir, sizeb, best_now, optional_data):
    logger.info('checkpoint', extra={
        'event_type': EventType.CHECKPOINT,
        'id': checkpoint_idx,
        'subdir': subdir,
        'sizeb': sizeb,
        'best_now': best_now,
        'optional': optional_data
    })



# coding: utf-8

from enum import Enum
import pandas as pd
import urllib

from supervisely_lib.api.module_api import ApiField, CloneableModuleApi, UpdateableModule, RemoveableModuleApi
from supervisely_lib.project.project_meta import ProjectMeta
from supervisely_lib.project.project_type import ProjectType


class ProjectNotFound(Exception):
    pass


class ExpectedProjectTypeMismatch(Exception):
    pass


class ProjectApi(CloneableModuleApi, UpdateableModule, RemoveableModuleApi):
    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.NAME,
                ApiField.DESCRIPTION,
                ApiField.SIZE,
                ApiField.README,
                ApiField.WORKSPACE_ID,
                ApiField.IMAGES_COUNT,  # for compatibility with existing code
                ApiField.ITEMS_COUNT,
                ApiField.DATASETS_COUNT,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT,
                ApiField.TYPE,
                ApiField.REFERENCE_IMAGE_URL]

    @staticmethod
    def info_tuple_name():
        return 'ProjectInfo'

    def __init__(self, api):
        CloneableModuleApi.__init__(self, api)
        UpdateableModule.__init__(self, api)

    def get_list(self, workspace_id, filters=None):
        '''
        :param workspace_id: int
        :param filters: list
        :return: list all the projects for a given workspace
        '''
        return self.get_list_all_pages('projects.list',  {ApiField.WORKSPACE_ID: workspace_id, "filter": filters or []})

    def get_info_by_id(self, id, expected_type=None, raise_error=False):
        '''
        :param id: int
        :param expected_type: type of data we expext to get info (raise error if type of project is not None and != expected type)
        :param raise_error: bool
        :return: project metadata by numeric id (None if request status_code == 404, raise error in over way)
        '''
        info = self._get_info_by_id(id, 'projects.info')
        self._check_project_info(info, id=id, expected_type=expected_type, raise_error=raise_error)
        return info

    def get_info_by_name(self, parent_id, name, expected_type=None, raise_error=False):
        '''
        :param parent_id: int
        :param name: str
        :param expected_type: type of data we expext to get info
        :param raise_error: bool
        :return: project metadata by numeric workspace id and given name of project
        '''
        info = super().get_info_by_name(parent_id, name)
        self._check_project_info(info, name=name, expected_type=expected_type, raise_error=raise_error)
        return info

    def _check_project_info(self, info, id=None, name=None, expected_type=None, raise_error=False):
        '''
        Checks if a project exists with a given id and type of project == expected type
        :param info: project metadata information
        :param id: int
        :param name: str
        :param expected_type: type of data we expext to get info
        :param raise_error: bool
        '''
        if raise_error is False:
            return

        str_id = ""
        if id is not None:
            str_id += "id: {!r} ".format(id)
        if name is not None:
            str_id += "name: {!r}".format(name)

        if info is None:
            raise ProjectNotFound("Project {} not found".format(str_id))
        if expected_type is not None and info.type != str(expected_type):
            raise ExpectedProjectTypeMismatch("Project {!r} has type {!r}, but expected type is {!r}"
                                              .format(str_id, info.type, expected_type))

    def get_meta(self, id):
        '''
        :param id: int
        :return: labeling meta information for the project - the set of available object classes and tags
        '''
        response = self._api.post('projects.meta', {'id': id})
        return response.json()

    def create(self, workspace_id, name, type=ProjectType.IMAGES, description="", change_name_if_conflict=False):
        '''
        Create project with given name in workspace with given id
        :param workspace_id: int
        :param name: str
        :param type: type of progect to create
        :param description: str
        :param change_name_if_conflict: bool
        :return: created project metadata
        '''
        effective_name = self._get_effective_new_name(
            parent_id=workspace_id, name=name, change_name_if_conflict=change_name_if_conflict)
        response = self._api.post('projects.add', {ApiField.WORKSPACE_ID: workspace_id,
                                                   ApiField.NAME: effective_name,
                                                   ApiField.DESCRIPTION: description,
                                                   ApiField.TYPE: str(type)})
        return self._convert_json_info(response.json())

    def _get_update_method(self):
        return 'projects.editInfo'

    def update_meta(self, id, meta):
        '''
        Update given project with given metadata
        :param id: int
        :param meta: project metainformation
        '''
        self._api.post('projects.meta.update', {ApiField.ID: id, ApiField.META: meta})

    def _clone_api_method_name(self):
        return 'projects.clone'

    def get_datasets_count(self, id):
        '''
        :param id: int
        :return: int (number of datasets in given project)
        '''
        datasets = self._api.dataset.get_list(id)
        return len(datasets)

    def get_images_count(self, id):
        '''
        :param id: int
        :return: int (number of images in given project)
        '''
        datasets = self._api.dataset.get_list(id)
        return sum([dataset.images_count for dataset in datasets])

    def _remove_api_method_name(self):
        return 'projects.remove'

    def merge_metas(self, src_project_id, dst_project_id):
        '''
        Add metadata from given progect to given destination project
        :param src_project_id: int
        :param dst_project_id: int
        :return: merged project metainformation
        '''
        if src_project_id == dst_project_id:
            return self.get_meta(src_project_id)

        src_meta = ProjectMeta.from_json(self.get_meta(src_project_id))
        dst_meta = ProjectMeta.from_json(self.get_meta(dst_project_id))

        new_dst_meta = src_meta.merge(dst_meta)
        new_dst_meta_json = new_dst_meta.to_json()
        self.update_meta(dst_project_id, new_dst_meta.to_json())

        return new_dst_meta_json

    def get_activity(self, id):
        #@TODO: reimplement - use api.team.get_activity with project_id filter
        response = self._api.post('projects.activity', {ApiField.ID: id})
        df = pd.DataFrame(response.json())
        return df

    def _convert_json_info(self, info: dict, skip_missing=True):
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        if res.reference_image_url is not None:
            res = res._replace(reference_image_url=urllib.parse.urljoin(self._api.server_address, res.reference_image_url))
        if res.items_count is None:
            res = res._replace(items_count=res.images_count)
        return res

    def get_stats(self, id):
        response = self._api.post('projects.stats', {ApiField.ID: id})
        return response.json()

    def url(self, id):
        result = urllib.parse.urljoin(self._api.server_address, 'projects/{}/datasets'.format(id))
        return result

    def update_custom_data(self, id, data):
        if type(data) is not dict:
            raise TypeError('Meta must be dict, not {!r}'.format(type(data)))
        response = self._api.post('projects.editInfo', {ApiField.ID: id, ApiField.CUSTOM_DATA: data})
        return response.json()



# coding: utf-8

import cv2
import numpy as np

from shapely.geometry import mapping, Polygon as ShapelyPolygon

from supervisely_lib.geometry.conversions import shapely_figure_to_coords_list
from supervisely_lib.geometry.point_location import row_col_list_to_points, points_to_row_col_list
from supervisely_lib.geometry.vector_geometry import VectorGeometry
from supervisely_lib.geometry.constants import EXTERIOR, INTERIOR, POINTS, LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID
from supervisely_lib.geometry import validation
from supervisely_lib.sly_logger import logger


class Polygon(VectorGeometry):
    '''
    This is a class for creating and using Polygon objects for Labels
    '''
    @staticmethod
    def geometry_name():
        return 'polygon'

    def __init__(self, exterior, interior,
                 sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        :param exterior: list of PointLocation objects, the object contour is defined with these points
        :param interior: list of elements that has the same structure like the "exterior" field. This is the list of polygons that define object holes.
        '''
        if len(exterior) < 3:
            raise ValueError('"{}" field must contain at least 3 points to create "Polygon" object.'.format(EXTERIOR))
        if any(len(element) < 3 for element in interior):
            raise ValueError('"{}" element must contain at least 3 points.'.format(INTERIOR))

        super().__init__(exterior, interior, sly_id=sly_id, class_id=class_id, labeler_login=labeler_login,
                         updated_at=updated_at, created_at=created_at)

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert Poligon from json format to Poligon class object. If json format is not correct it generate exception error.
        :param data: input Poligon in json format
        :return: Poligon class object
        '''
        validation.validate_geometry_points_fields(data)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(exterior=row_col_list_to_points(data[POINTS][EXTERIOR], flip_row_col_order=True),
                   interior=[row_col_list_to_points(i, flip_row_col_order=True) for i in data[POINTS][INTERIOR]],
                   sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    def crop(self, rect):
        '''
        Crop the current Polygon with a given rectangle, if polygon cat't be cropped it generate exception error
        :param rect: Rectangle class object
        :return: list of Poligon class objects
        '''
        from supervisely_lib.geometry.point_location import PointLocation
        try:
            points = [PointLocation(row=rect.top, col=rect.left), PointLocation(row=rect.top, col=rect.right + 1),
                PointLocation(row=rect.bottom + 1, col=rect.right + 1), PointLocation(row=rect.bottom + 1, col=rect.left)]
            #points = rect.corners # old implementation with 1 pixel error (right bottom) #@TODO: investigate here (critical issue)

            clipping_window_shpl = ShapelyPolygon(points_to_row_col_list(points))
            self_shpl = ShapelyPolygon(self.exterior_np, holes=self.interior_np)
            intersections_shpl = self_shpl.buffer(0).intersection(clipping_window_shpl)
            mapping_shpl = mapping(intersections_shpl)
        except Exception:
            logger.warn('Polygon cropping exception, shapely.', exc_info=True)
            # raise
            # if polygon is invalid, just print warning and skip it
            # @TODO: need more investigation here
            return []

        intersections = shapely_figure_to_coords_list(mapping_shpl)

        # Check for bad cropping cases (e.g. empty points list)
        out_polygons = []
        for intersection in intersections:
            if isinstance(intersection, list) and len(intersection) > 0 and len(intersection[0]) >= 3:
                exterior = row_col_list_to_points(intersection[0], do_round=True)
                interiors = []
                for interior_contour in intersection[1:]:
                    if len(interior_contour) > 2:
                        interiors.append(row_col_list_to_points(interior_contour, do_round=True))
                out_polygons.append(Polygon(exterior, interiors))
        return out_polygons

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        exterior = self.exterior_np[:, ::-1]
        interior = [x[:, ::-1] for x in self.interior_np]
        bmp_to_draw = np.zeros(bitmap.shape[:2], np.uint8)
        cv2.fillPoly(bmp_to_draw, pts=[exterior], color=1)
        cv2.fillPoly(bmp_to_draw, pts=interior, color=0)
        bool_mask = bmp_to_draw.astype(bool)
        bitmap[bool_mask] = color

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        exterior = self.exterior_np[:, ::-1]
        interior = [x[:, ::-1] for x in self.interior_np]

        poly_lines = [exterior] + interior
        cv2.polylines(bitmap, pts=poly_lines, isClosed=True, color=color, thickness=thickness)

    # @TODO: extend possibilities, consider interior
    # returns area of exterior figure only
    @property
    def area(self):
        '''
        :return: area of current Poligon(exterior figure only)
        '''
        exterior = self.exterior_np
        return self._get_area_by_gauss_formula(exterior[:, 0], exterior[:, 1])

    @staticmethod
    def _get_area_by_gauss_formula(rows, cols):
        return 0.5 * np.abs(np.dot(rows, np.roll(cols, 1)) - np.dot(cols, np.roll(rows, 1)))

    def approx_dp(self, epsilon):
        '''
        The function approx_dp approximates a polygonal curve with the specified precision
        :param epsilon: Parameter specifying the approximation accuracy. This is the maximum distance between the original curve and its approximation.
        :return: Poligon class object
        '''
        exterior_np = self._approx_ring_dp(self.exterior_np, epsilon, closed=True).tolist()
        interior_np = [self._approx_ring_dp(x, epsilon, closed=True).tolist() for x in self.interior_np]
        exterior = row_col_list_to_points(exterior_np, do_round=True)
        interior = [row_col_list_to_points(x, do_round=True) for x in interior_np]
        return Polygon(exterior, interior)

    @classmethod
    def allowed_transforms(cls):
        from supervisely_lib.geometry.any_geometry import AnyGeometry
        from supervisely_lib.geometry.rectangle import Rectangle
        from supervisely_lib.geometry.bitmap import Bitmap
        return [AnyGeometry, Rectangle, Bitmap]



# coding: utf-8

import os
import skvideo.io
from supervisely_lib import logger as default_logger
from supervisely_lib.io.fs import get_file_name, get_file_ext
from supervisely_lib._utils import rand_str


# Do NOT use directly for video extension validation. Use is_valid_ext() /  has_valid_ext() below instead.
ALLOWED_VIDEO_EXTENSIONS = ['.avi', '.mp4', '.3gp', '.flv', '.webm', '.wmv', '.mov', '.mkv']


_SUPPORTED_CONTAINERS = {'mp4', 'webm', 'ogg', 'ogv'}
_SUPPORTED_CODECS = {'h264', 'vp8', 'vp9'}


class VideoExtensionError(Exception):
    pass


class UnsupportedVideoFormat(Exception):
    pass


class VideoReadException(Exception):
    pass


def is_valid_ext(ext: str) -> bool:
    '''
    Checks if given extention is supported
    :param ext: str
    :return: bool
    '''
    return ext.lower() in ALLOWED_VIDEO_EXTENSIONS


def has_valid_ext(path: str) -> bool:
    '''
    Checks if file from given path with given extention is supported
    :param path: str
    :return: bool
    '''
    return is_valid_ext(os.path.splitext(path)[1])


def validate_ext(ext: str):
    '''
    Raise error if given extention is not supported
    :param ext: str
    '''
    if not is_valid_ext(ext):
        raise UnsupportedVideoFormat('Unsupported video extension: {}. Only the following extensions are supported: {}.'
                                     .format(ext, ALLOWED_VIDEO_EXTENSIONS))


def get_image_size_and_frames_count(path):
    '''
    Find size of image and number of frames from given video path
    :param path: str
    :return: tuple of integers, int
    '''
    vreader = skvideo.io.FFmpegReader(path)
    vlength = vreader.getShape()[0]
    img_height = vreader.getShape()[1]
    img_width = vreader.getShape()[2]

    img_size = (img_height, img_width)

    return img_size, vlength


def validate_format(path):
    '''
    Raise error if video from given path can't be read or file extention from given path with is not supported
    :param path: str
    '''
    try:
        get_image_size_and_frames_count(path)
    except Exception as e:
        raise VideoReadException(
            'Error has occured trying to read video {!r}. Original exception message: {!r}'.format(path, str(e)))

    validate_ext(os.path.splitext(path)[1])


def _check_video_requires_processing(video_info, stream_info):
    '''
    Check if video need container or codec processing
    :param video_info: dict
    :param stream_info: dict
    :return: bool
    '''
    need_process_container = True
    for name in video_info["meta"]["formatName"].split(','):
        name = name.strip().split('.')[-1]
        if name in _SUPPORTED_CONTAINERS:
            need_process_container = False
            break

    need_process_codec = True
    codec = stream_info["codecName"]
    if codec in _SUPPORTED_CODECS:
        need_process_codec = False

    if (need_process_container is False) and (need_process_codec is False):
        return False

    return True


def count_video_streams(all_streams):
    '''
    Count number of video streams
    :param all_streams: list of streams(dict)
    :return: int
    '''
    count = 0
    for stream_info in all_streams:
        if stream_info["codecType"] == "video":
            count += 1
    return count


def get_video_streams(all_streams):
    '''
    Get list of video streams from given list of all streams
    :param all_streams: list of streams(dict)
    :return: list
    '''
    video_streams = []
    for stream_info in all_streams:
        if stream_info["codecType"] == "video":
            video_streams.append(stream_info)
    return video_streams


def warn_video_requires_processing(file_name, logger=None):
    '''
    Create logger if it was not there and displays message about the need for transcoding
    :param file_name: str
    :param logger: logger class object
    '''
    if logger is None:
        logger = default_logger
    logger.warning("Video Stream {!r} is skipped: requires transcoding. Transcoding is supported only in Enterprise Edition (EE)".format(file_name))


def gen_video_stream_name(file_name, stream_index):
    '''
    Create name to video stream from given filename and index of stream
    :param file_name: str
    :param stream_index: int
    :return: str
    '''
    return "{}_stream_{}_{}{}".format(get_file_name(file_name), stream_index, rand_str(5), get_file_ext(file_name))



# coding: utf-8
import uuid

from supervisely_lib.api.api import Api
from supervisely_lib.api.module_api import ApiField
from supervisely_lib.app_widget.widget_type import WidgetType
from supervisely_lib.app_widget.base_widget import BaseWidget


class GalleryWidget(BaseWidget):
    widget_type = WidgetType.GALLERY

    @classmethod
    def create(cls, name, description, image_mask_pairs=None, id=None):
        res = dict()
        res[ApiField.TYPE] = str(cls.widget_type)
        res[ApiField.NAME] = name
        res[ApiField.SUBTITLE] = description
        res[ApiField.CONTENT] = image_mask_pairs if image_mask_pairs is not None else []
        res[ApiField.ID] = uuid.uuid4().hex if id is None else id
        return res

    @classmethod
    def add_item(cls, api: Api, app_id, widget_id, img_url, mask_url):
        item = [[img_url, mask_url]]
        return api.report.update_widget(report_id=app_id, widget_id=widget_id, content=item)

    @classmethod
    def replace_items(cls, api: Api, app_id, widget_id, items):
        return api.report.rewrite_widget(report_id=app_id, widget_id=widget_id, widget_type=str(cls.widget_type), content=items)


# coding: utf-8

import requests
from collections import namedtuple
from copy import deepcopy
from supervisely_lib._utils import batched

from supervisely_lib._utils import camel_to_snake


class ApiField:
    ID =                'id'
    NAME =              'name'
    DESCRIPTION =       'description'
    CREATED_AT =        'createdAt'
    UPDATED_AT =        'updatedAt'
    ROLE =              'role'
    TEAM_ID =           'teamId'
    WORKSPACE_ID =      'workspaceId'
    CONFIG =            'config'
    SIZE =              'size'
    PLUGIN_ID =         'pluginId'
    PLUGIN_VERSION =    'pluginVersion'
    HASH =              'hash'
    STATUS =            'status'
    ONLY_TRAIN =        'onlyTrain'
    USER_ID =           'userId'
    README =            'readme'
    PROJECT_ID =        'projectId'
    IMAGES_COUNT =      'imagesCount'
    ITEMS_COUNT =       'itemsCount'
    DATASET_ID =        'datasetId'
    LINK =              'link'
    WIDTH =             'width'
    HEIGHT =            'height'
    WEIGHTS_LOCATION =  'weightsLocation'
    IMAGE_ID =          'imageId'
    IMAGE_NAME =        'imageName'
    ANNOTATION =        'annotation'
    TASK_ID =           'taskId'
    LABELS_COUNT =      'labelsCount'
    FILTER =            'filter'
    META =              'meta'
    SHARED_LINK =       'sharedLinkToken'
    EXPLORE_PATH =      'explorePath'
    MIME =              'mime'
    EXT =               'ext'
    TYPE =              'type'
    DEFAULT_VERSION =   'defaultVersion'
    DOCKER_IMAGE =      'dockerImage'
    CONFIGS =           'configs'
    VERSIONS =          'versions'
    VERSION =           'version'
    TOKEN =             'token'
    CAPABILITIES =      'capabilities'
    STARTED_AT =        'startedAt'
    FINISHED_AT =       'finishedAt'
    AGENT_ID =          'agentId'
    MODEL_ID =          'modelId'
    RESTART_POLICY =    'restartPolicy'
    SETTINGS =          'settings'
    SORT =              'sort'
    SORT_ORDER =        'sort_order'
    IMAGES =            'images'
    IMAGE_IDS =         'imageIds'
    ANNOTATIONS =       'annotations'
    EMAIL =             'email'
    LOGIN =             'login'
    LOGINS =            'logins'
    DISABLED =          'disabled'
    LAST_LOGIN =        'lastLogin'
    PASSWORD =          'password'
    ROLE_ID =           'roleId'
    IS_RESTRICTED =     'isRestricted'
    DISABLE =           'disable'
    TEAMS =             'teams'
    USER_IDS =          'userIds'
    PROJECT_NAME =      (['projectTitle'], 'project_name')
    DATASET_NAME =      (['datasetTitle'], 'dataset_name')
    WORKSPACE_NAME =    (['workspaceTitle'], 'workspace_name')
    CREATED_BY_ID =     (['createdBy'], 'created_by_id')
    CREATED_BY_LOGIN =  (['managerLogin'], 'created_by_login')
    ASSIGNED_TO_ID =    (['userId'], 'assigned_to_id')
    ASSIGNED_TO_LOGIN = (['labelerLogin'], 'assigned_to_login')
    FINISHED_IMAGES_COUNT = 'finishedImagesCount'
    PROGRESS_IMAGES_COUNT = 'progressImagesCount'
    CLASSES_TO_LABEL =  (['meta', 'classes'], 'classes_to_label')
    TAGS_TO_LABEL =     (['meta', 'projectTags'], 'tags_to_label')
    IMAGES_RANGE =      (['meta', 'range'], 'images_range')
    REJECTED_IMAGES_COUNT = 'rejectedImagesCount'
    ACCEPTED_IMAGES_COUNT = 'acceptedImagesCount'
    OBJECTS_LIMIT_PER_IMAGE = (['meta', 'imageFiguresLimit'], 'objects_limit_per_image')
    TAGS_LIMIT_PER_IMAGE = (['meta', 'imageTagsLimit'], 'tags_limit_per_image')
    FILTER_IMAGES_BY_TAGS = (['meta', 'imageTags'], 'filter_images_by_tags')
    INCLUDE_IMAGES_WITH_TAGS = ([], 'include_images_with_tags')
    EXCLUDE_IMAGES_WITH_TAGS = ([], 'exclude_images_with_tags')
    SIZEB =             (['size'], 'sizeb')
    FRAMES_COUNT =      (['fileMeta', 'framesCount'], 'frames_count')
    PATH_ORIGINAL =     'pathOriginal'
    OBJECTS_COUNT =     'objectsCount'
    FRAMES_TO_TIMECODES = (['fileMeta', 'framesToTimecodes'], 'frames_to_timecodes')
    TAGS =              'tags'
    VIDEO_ID =          'videoId'
    FRAME_INDEX =       (['meta', 'frame'], 'frame_index')
    LABELING_TOOL =     'tool'
    GEOMETRY_TYPE =     'geometryType'
    GEOMETRY =          'geometry'
    OBJECT_ID =         'objectId'
    FRAME =             'frame'
    #FIGURES_COUNT =     (['labelsCount'], 'figures_count')
    STREAMS =           'streams'
    VIDEO_IDS =         'videoIds'
    FRAME_WIDTH =       (['fileMeta', 'width'], 'frame_width')
    FRAME_HEIGHT =      (['fileMeta', 'height'], 'frame_height')
    VIDEO_NAME =        'videoName'
    FRAME_RANGE =       'frameRange'
    TRACK_ID =          'trackId'
    PROGRESS =          'progress'
    CURRENT =           'current'
    TOTAL =             'total'
    STOPPED =           'stopped'
    VIDEOS =            'videos'
    FILENAME =          'filename'
    SHAPE =             'shape'
    COLOR =             'color'
    CLASS_ID =          'classId'
    ENTITY_ID =         'entityId'
    ANNOTATION_OBJECTS = 'annotationObjects'
    TAG_ID =            'tagId'
    ERROR =             'error'
    MESSAGE =           'message'
    CONTENT =           'content'
    FIGURES =           'figures'
    LAYOUT =            'layout'
    WIDGETS =           'widgets'
    CLOUD_MIME =        (['fileMeta', 'mime'], 'cloud_mime')
    PREVIEW =           'preview'
    FIGURES_COUNT =     'figuresCount'
    ANN_OBJECTS_COUNT =     (['annotationObjectsCount'], 'objects_count')
    POINTCLOUD_ID =     'pointCloudId'
    POINTCLOUD_IDS =    'pointCloudIds'
    POINTCLOUDS =       'pointClouds'
    ADVANCED =          'advanced'
    IGNORE_AGENT =      'ignoreAgent'
    SCRIPT =            'script'
    LOGS =              'logs'
    FILES =             'files'
    HASHES =            'hashes'
    SUBTITLE =          'subtitle'
    COMMAND =           'command'
    DEFAULT_VALUE =     'defaultValue'
    TITLE =             'title'
    AREA =              'area'
    OPTIONS =           'options'
    REPORT_ID =         'reportId'
    WIDGET =            'widget'
    LAYOUT =            'layout'
    PAYLOAD =           'payload'
    FIELD =             'field'
    FIELDS =            'fields'
    APPEND =            'append'
    WITH_CUSTOM_DATA =  'withCustomBigData'
    PATH =              'path'
    SESSION_ID =        'sessionId'
    ACTION =            'action'
    FIGURE_ID =         'figureId'
    VALUE =             'value'
    ZOOM_FACTOR =       'zoomFactor'
    FULL_STORAGE_URL =  'fullStorageUrl'
    REVIEWER_ID =       'reviewerId'
    REVIEWER_LOGIN =    'reviewerLogin'
    RECURSIVE =         'recursive'
    ECOSYSTEM_ITEM_ID = 'moduleId'
    APP_ID =            'appId'
    PROJECT =           'project'
    OUTPUT =            'output'
    REFERENCE_IMAGE_URL='referenceImageUrl'
    GENERAL =           'general'
    ENTITIES =          'entities'
    STORAGE_PATH =      'storagePath'
    EXT2 =              (['meta', 'ext'], 'ext')
    MIME2 =             (['meta', 'mime'], 'mime')
    SIZEB2 =            (['meta', 'size'], 'sizeb')
    JOB_ID =            'jobId'
    DATASETS_COUNT =    'datasetsCount'
    CUSTOM_DATA =       'customData'
    CONTEXT =           'context'
    STATE =             'state'


def _get_single_item(items):
    if len(items) == 0:
        return None
    if len(items) > 1:
        raise RuntimeError('There are several items with the same name')
    return items[0]


class _JsonConvertibleModule:
    def _convert_json_info(self, info: dict, skip_missing=False):
        raise NotImplementedError()


class ModuleApiBase(_JsonConvertibleModule):
    MAX_WAIT_ATTEMPTS = 999
    WAIT_ATTEMPT_TIMEOUT_SEC = 1

    @staticmethod
    def info_sequence():
        raise NotImplementedError()

    @staticmethod
    def info_tuple_name():
        raise NotImplementedError()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        try:
            field_names = []
            for name in cls.info_sequence():
                if type(name) is str:
                    field_names.append(camel_to_snake(name))
                elif type(name) is tuple and type(name[1]) is str:
                    field_names.append(name[1])
                else:
                    raise RuntimeError('Can not parse field {!r}'.format(name))
            cls.InfoType = namedtuple(cls.info_tuple_name(), field_names)
        except NotImplementedError:
            pass

    def __init__(self, api):
        self._api = api

    def _add_sort_param(self, data):
        results = deepcopy(data)
        results[ApiField.SORT] = ApiField.ID
        results[ApiField.SORT_ORDER] = 'asc'  # @TODO: move to enum
        return results

    def get_list_all_pages(self, method, data, progress_cb=None, convert_json_info_cb=None):
        if convert_json_info_cb is None:
            convert_func = self._convert_json_info
        else:
            convert_func = convert_json_info_cb

        if ApiField.SORT not in data:
            data = self._add_sort_param(data)
        first_response = self._api.post(method, data).json()
        total = first_response['total']
        per_page = first_response['perPage']
        pages_count = first_response['pagesCount']

        results = first_response['entities']
        if progress_cb is not None:
            progress_cb(len(results))
        if pages_count == 1 and len(first_response['entities']) == total:
            pass
        else:
            for page_idx in range(2, pages_count + 1):
                temp_resp = self._api.post(method, {**data, 'page': page_idx, 'per_page': per_page})
                temp_items = temp_resp.json()['entities']
                results.extend(temp_items)
                if progress_cb is not None:
                    progress_cb(len(temp_items))
            if len(results) != total:
                raise RuntimeError('Method {!r}: error during pagination, some items are missed'.format(method))

        return [convert_func(item) for item in results]

    @staticmethod
    def _get_info_by_name(get_info_by_filters_fn, name):
        filters = [{"field": ApiField.NAME, "operator": "=", "value": name}]
        return get_info_by_filters_fn(filters)

    def get_info_by_id(self, id):
        raise NotImplementedError()

    @staticmethod
    def _get_free_name(exist_check_fn, name):
        res_title = name
        suffix = 1
        while exist_check_fn(res_title):
            res_title = '{}_{:03d}'.format(name, suffix)
            suffix += 1
        return res_title

    def _convert_json_info(self, info: dict, skip_missing=False):
        def _get_value(dict, field_name, skip_missing):
            if skip_missing is True:
                return dict.get(field_name, None)
            else:
                return dict[field_name]

        if info is None:
            return None
        else:
            field_values = []
            for field_name in self.info_sequence():
                if type(field_name) is str:
                    field_values.append(_get_value(info, field_name, skip_missing))
                elif type(field_name) is tuple:
                    value = None
                    for sub_name in field_name[0]:
                        if value is None:
                            value = _get_value(info, sub_name, skip_missing)
                        else:
                            value = _get_value(value, sub_name, skip_missing)
                    field_values.append(value)
                else:
                    raise RuntimeError('Can not parse field {!r}'.format(field_name))
            return self.InfoType(*field_values)

    def _get_response_by_id(self, id, method, id_field):
        try:
            return self._api.post(method, {id_field: id})
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                return None
            else:
                raise error

    def _get_info_by_id(self, id, method):
        response = self._get_response_by_id(id, method, id_field=ApiField.ID)
        return self._convert_json_info(response.json()) if (response is not None) else None


# Base class for entities that have a parent object in the system.
class ModuleApi(ModuleApiBase):
    def __init(self, api):
        self._api = api

    def get_info_by_name(self, parent_id, name):
        return self._get_info_by_name(
            get_info_by_filters_fn=lambda module_name: self._get_info_by_filters(parent_id, module_name), name=name)

    def _get_info_by_filters(self, parent_id, filters):
        items = self.get_list(parent_id, filters)
        return _get_single_item(items)

    def get_list(self, parent_id, filters=None):
        raise NotImplementedError()

    def exists(self, parent_id, name):
        return self.get_info_by_name(parent_id, name) is not None

    def get_free_name(self, parent_id, name):
        return self._get_free_name(exist_check_fn=lambda module_name: self.exists(parent_id, module_name), name=name)

    def _get_effective_new_name(self, parent_id, name, change_name_if_conflict=False):
        return self.get_free_name(parent_id, name) if change_name_if_conflict else name


# Base class for entities that do not have a parent object in the system.
class ModuleNoParent(ModuleApiBase):
    def get_info_by_name(self, name):
        return self._get_info_by_name(get_info_by_filters_fn=self._get_info_by_filters, name=name)

    def _get_info_by_filters(self, filters):
        items = self.get_list(filters)
        return _get_single_item(items)

    def get_list(self, filters=None):
        raise NotImplementedError()

    def exists(self, name):
        return self.get_info_by_name(name) is not None

    def get_free_name(self, name):
        return self._get_free_name(exist_check_fn=lambda module_name: self.exists(module_name), name=name)

    def _get_effective_new_name(self, name, change_name_if_conflict=False):
        return self.get_free_name(name) if change_name_if_conflict else name


class CloneableModuleApi(ModuleApi):
    def _clone_api_method_name(self):
        raise NotImplementedError()

    def _clone(self, clone_type: dict, dst_workspace_id: int, dst_name: str):
        response = self._api.post(self._clone_api_method_name(), {**clone_type,
                                                                  ApiField.WORKSPACE_ID: dst_workspace_id,
                                                                  ApiField.NAME: dst_name})
        return response.json()[ApiField.TASK_ID]

    def clone(self, id, dst_workspace_id, dst_name):
        return self._clone({ApiField.ID: id}, dst_workspace_id, dst_name)

    def clone_by_shared_link(self, shared_link, dst_workspace_id, dst_name):
        return self._clone({ApiField.SHARED_LINK: shared_link}, dst_workspace_id, dst_name)

    def clone_from_explore(self, explore_path, dst_workspace_id, dst_name):
        return self._clone({ApiField.EXPLORE_PATH: explore_path}, dst_workspace_id, dst_name)

    def get_or_clone_from_explore(self, explore_path, dst_workspace_id, dst_name):
        if not self.exists(dst_workspace_id, dst_name):
            task_id = self.clone_from_explore(explore_path, dst_workspace_id, dst_name)
            self._api.task.wait(task_id, self._api.task.Status.FINISHED)
        item = self.get_info_by_name(dst_workspace_id, dst_name)
        return item


class ModuleWithStatus:
    def get_status(self, id):
        raise NotImplementedError()

    def raise_for_status(self, status):
        raise NotImplementedError()


class WaitingTimeExceeded(Exception):
    pass


class UpdateableModule(_JsonConvertibleModule):
    def __init__(self, api):
        self._api = api

    def _get_update_method(self):
        raise NotImplementedError()

    def update(self, id, name=None, description=None):
        if name is None and description is None:
            raise ValueError("\'name\' or \'description\' or both have to be specified")

        body = {ApiField.ID: id}
        if name is not None:
            body[ApiField.NAME] = name
        if description is not None:
            body[ApiField.DESCRIPTION] = description

        response = self._api.post(self._get_update_method(), body)
        return self._convert_json_info(response.json())


class RemoveableModuleApi(ModuleApi):
    def _remove_api_method_name(self):
        raise NotImplementedError()

    def remove(self, id):
        self._api.post(self._remove_api_method_name(), {ApiField.ID: id})

    def remove_batch(self, ids, progress_cb=None):
        for id in ids:
            self.remove(id)
            if progress_cb is not None:
                progress_cb(1)


class RemoveableBulkModuleApi(ModuleApi):
    def _remove_batch_api_method_name(self):
        raise NotImplementedError()

    def _remove_batch_field_name(self):
        raise NotImplementedError()

    def remove_batch(self, ids, progress_cb=None):
        for ids_batch in batched(ids):
            self._api.post(self._remove_batch_api_method_name(), {self._remove_batch_field_name(): ids_batch})
            if progress_cb is not None:
                progress_cb(len(ids_batch))

    def remove(self, id):
        self.remove_batch([id])



# coding: utf-8

import os


# Do NOT use directly for extension validation. Use is_valid_ext() /  has_valid_ext() below instead.
ALLOWED_POINTCLOUD_EXTENSIONS = ['.pcd']


class PointcloudExtensionError(Exception):
    pass


class UnsupportedPointcloudFormat(Exception):
    pass


class PointcloudReadException(Exception):
    pass


def is_valid_ext(ext: str) -> bool:
    '''
    Checks if given extention is supported
    :param ext: str
    :return: bool
    '''
    return ext.lower() in ALLOWED_POINTCLOUD_EXTENSIONS


def has_valid_ext(path: str) -> bool:
    '''
    Checks if file from given path with given extention is supported
    :param path: str
    :return: bool
    '''
    return is_valid_ext(os.path.splitext(path)[1])


def validate_ext(ext: str):
    '''
    Raise error if given extention is not supported
    :param ext: str
    '''
    if not is_valid_ext(ext):
        raise UnsupportedPointcloudFormat('Unsupported pointcloud extension: {}. Only the following extensions are supported: {}.'
                                          .format(ext, ALLOWED_POINTCLOUD_EXTENSIONS))


def validate_format(path):
    #@TODO: later
    validate_ext(path)



# coding: utf-8
import os.path

import torch
from torch.nn import Parameter


class WeightsRW:
    """
    Help to load weights for PyTorch model from file and save current model state dict into given place.

    Args:
        model_dir: Path to the folder for storing weights.
        model_file: Name of weights file(default: model.pt).
    """
    def __init__(self, model_dir, model_file=None):
        self._weights_fpath = os.path.join(model_dir, model_file or 'model.pt')

    def save(self, model):
        torch.save(model.state_dict(), self._weights_fpath)

    @staticmethod
    def _transfer_params(src_state, dest_model):
        """Copies parameters and buffers from :attr:`src_state` into
        :attr:`dest_model` module and its descendants.

        Arguments:
            src_state (dict): A dict containing parameters and
                persistent buffers.
            dest_model: model
        """

        dest_state = dest_model.state_dict()
        for name, param in src_state.items():
            if name in dest_state:
                if isinstance(param, Parameter):
                    # backwards compatibility for serialized parameters
                    param = param.data
                try:
                    dest_state[name].copy_(param)
                except Exception:
                    raise RuntimeError('While copying the parameter named {}, '
                                       'whose dimensions in the model are {} and '
                                       'whose dimensions in the checkpoint are {}.'
                                       .format(name, dest_state[name].size(), param.size()))

    def load_for_transfer_learning(self, model, ignore_matching_layers=None, logger=None):
        ignore_matching_layers = ignore_matching_layers or []
        snapshot_weights = torch.load(self._weights_fpath)

        # Remove the layers matching the requested patterns (usually done for the head layers in transfer learning).
        # Make an explicit set for easier logging.
        to_delete = set(el for el in snapshot_weights.keys() if
                        any(delete_substring in el for delete_substring in ignore_matching_layers))
        snapshot_weights = {k: v for k, v in snapshot_weights.items() if k not in to_delete}
        if len(to_delete) > 0 and logger is not None:
            logger.info('Skip weight init for output layers.', extra={'layer_names': sorted(to_delete)})

        self._transfer_params(snapshot_weights, model)
        return model

    def load_strictly(self, model):
        snapshot_weights = torch.load(self._weights_fpath)

        # Make sure the sets of parameters exactly match between the model and the snapshot.
        snapshot_keys = set(snapshot_weights.keys())
        model_keys = set(model.state_dict().keys())
        snapshot_extra_keys = snapshot_keys - model_keys
        if len(snapshot_extra_keys) > 0:
            raise KeyError('Parameters found in the snapshot file, but not in the model: {}'.format(snapshot_extra_keys))
        model_extra_keys = model_keys - snapshot_keys
        if len(model_extra_keys) > 0:
            raise KeyError('Model parameters missing from the snapshot file: {}'.format(model_extra_keys))

        self._transfer_params(snapshot_weights, model)
        return model



# coding: utf-8

from supervisely_lib.io.json import JsonSerializable
from supervisely_lib.annotation.obj_class_collection import ObjClassCollection
from supervisely_lib.annotation.tag_meta_collection import TagMetaCollection
from supervisely_lib._utils import take_with_default


class ProjectMetaJsonFields:
    OBJ_CLASSES = 'classes'
    IMG_TAGS = 'tags_images'
    OBJ_TAGS = 'tags_objects'
    TAGS = 'tags'
    PROJECT_TYPE = 'projectType'


def _merge_img_obj_tag_metas(img_tag_metas: ObjClassCollection,
                             obj_tag_metas: ObjClassCollection) -> ObjClassCollection:
    obj_tag_metas_to_add = []
    for obj_tag_meta in obj_tag_metas:
        img_tag_meta_same_key = img_tag_metas.get(obj_tag_meta.key(), None)
        if img_tag_meta_same_key is None:
            obj_tag_metas_to_add.append(obj_tag_meta)
        elif not img_tag_meta_same_key.is_compatible(obj_tag_meta):
            raise ValueError(
                'Unable to merge tag metas for images and objects. Found tags with the same name, but incompatible '
                'values. \n Image-level tag meta: {}\n Object-level tag meta: {}.\n Rename one of the tags to have a '
                'unique name to be able to load project meta.'.format(str(img_tag_meta_same_key), str(obj_tag_meta)))
    return img_tag_metas.add_items(obj_tag_metas_to_add)


class ProjectMeta(JsonSerializable):
    '''
    This is a class for creating and using ProjectMeta objects. This class contain data about meta information of the project
    '''
    def __init__(self, obj_classes=None, tag_metas=None, project_type=None):
        '''
        :param obj_classes: Collection that stores ObjClass instances with unique names.
        :param tag_metas: Collection that stores TagMeta instances with unique names.
        '''
        self._obj_classes = ObjClassCollection() if obj_classes is None else obj_classes
        self._tag_metas = take_with_default(tag_metas, TagMetaCollection())
        self._project_type = project_type

    @property
    def obj_classes(self) -> ObjClassCollection:
        return self._obj_classes

    @property
    def tag_metas(self) -> TagMetaCollection:
        return self._tag_metas

    @property
    def project_type(self):
        return self._project_type

    def to_json(self):
        '''
        The function to_json convert ProjectMeta class object to json format
        :return: ProjectMeta in json format(dict)
        '''
        res = {
            ProjectMetaJsonFields.OBJ_CLASSES: self._obj_classes.to_json(),
            ProjectMetaJsonFields.TAGS: self._tag_metas.to_json()
        }
        if self._project_type is not None:
            res[ProjectMetaJsonFields.PROJECT_TYPE] = self._project_type
        return res

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert ProjectMeta from json format to ProjectMeta class object. Generate exception error if all project tags not in a single collection
        :param data: input ProjectMeta in json format
        :return: ProjectMeta class object
        '''
        tag_metas_json = data.get(ProjectMetaJsonFields.TAGS, [])
        img_tag_metas_json = data.get(ProjectMetaJsonFields.IMG_TAGS, [])
        obj_tag_metas_json = data.get(ProjectMetaJsonFields.OBJ_TAGS, [])
        project_type = data.get(ProjectMetaJsonFields.PROJECT_TYPE, None)

        if len(tag_metas_json) > 0:
            # New format - all project tags in a single collection.
            if any(len(x) > 0 for x in [img_tag_metas_json, obj_tag_metas_json]):
                raise ValueError(
                    'Project meta JSON contains both the {!r} section (current format merged tags for all of '
                    'the project) and {!r} or {!r} sections (legacy format with separate collections for images '
                    'and labeled objects). Either new format only or legacy format only are supported, but not a '
                    'mix.'.format(
                        ProjectMetaJsonFields.TAGS, ProjectMetaJsonFields.IMG_TAGS, ProjectMetaJsonFields.OBJ_TAGS))
            tag_metas = TagMetaCollection.from_json(tag_metas_json)
        else:
            img_tag_metas = TagMetaCollection.from_json(img_tag_metas_json)
            obj_tag_metas = TagMetaCollection.from_json(obj_tag_metas_json)
            tag_metas = _merge_img_obj_tag_metas(img_tag_metas, obj_tag_metas)

        return cls(obj_classes=ObjClassCollection.from_json(data[ProjectMetaJsonFields.OBJ_CLASSES]),
                   tag_metas=tag_metas, project_type=project_type)

    def merge(self, other):
        '''
        Merge all instances from given meta to ProjectMeta and return it copy
        :param other: ProjectMeta class object
        :return: ProjectMeta class object
        '''
        return self.clone(obj_classes=self._obj_classes.merge(other.obj_classes),
                          tag_metas=self._tag_metas.merge(other._tag_metas))

    def clone(self, obj_classes: ObjClassCollection = None, tag_metas: TagMetaCollection = None, project_type=None):
        '''
        The function clone create copy of ProjectMeta with given Collections that stores ObjClass and TagMeta
        :param obj_classes: ObjClassCollection class object
        :param tag_metas: TagMetaCollection class object
        :return: ProjectMeta class object
        '''
        return ProjectMeta(obj_classes=take_with_default(obj_classes, self.obj_classes),
                           tag_metas=take_with_default(tag_metas, self.tag_metas),
                           project_type=take_with_default(project_type, self.project_type))

    def add_obj_class(self, new_obj_class):
        '''
        The function add_obj_class add given objclass to ProjectMeta collection that stores ObjClass instances and return copy of ProjectMeta
        :param new_obj_class: ObjClass class object
        :return: ProjectMeta class object
        '''
        return self.add_obj_classes([new_obj_class])

    def add_obj_classes(self, new_obj_classes):
        '''
        The function add_obj_class add given objclasses to ProjectMeta collection that stores ObjClass instances and return copy of ProjectMeta
        :param new_obj_classes: list of ObjClass class objects
        :return: ProjectMeta class object
        '''
        return self.clone(obj_classes=self.obj_classes.add_items(new_obj_classes))

    def add_tag_meta(self, new_tag_meta):
        '''
        The function add_tag_meta add given tag to ProjectMeta collection that stores TagMeta instances and return copy of ProjectMeta
        :param new_tag_meta: TagMeta class object
        :return: ProjectMeta class object
        '''
        return self.add_tag_metas([new_tag_meta])

    def add_tag_metas(self, new_tag_metas):
        '''
        The function add_tag_metas add given tags to ProjectMeta collection that stores TagMeta instances and return copy of ProjectMeta
        :param new_tag_metas: list of TagMeta class objects
        :return: ProjectMeta class object
        '''
        return self.clone(tag_metas=self.tag_metas.add_items(new_tag_metas))

    @staticmethod
    def _delete_items(collection, item_names):
        '''
        :param collection: ObjClassCollection or TagMetaCollection instance
        :param item_names: list of item names to delete
        :return: list of items, which are in collection and not in given list of items to delete
        '''
        names_to_delete = set(item_names)
        res_items = []
        for item in collection:
            if item.key() not in names_to_delete:
                res_items.append(item)
        return res_items

    def delete_obj_class(self, obj_class_name):
        '''
        The function delete_obj_class delete objclass with given name from ProjectMeta collection that stores ObjClass instances and return copy of ProjectMeta
        :param obj_class_name: str(name of ObjClass to detele from collection)
        :return: ProjectMeta class object
        '''
        return self.delete_obj_classes([obj_class_name])

    def delete_obj_classes(self, obj_class_names):
        '''
        The function delete_obj_classes delete objclasses with given list of names from ProjectMeta collection that stores ObjClass instances and return copy of ProjectMeta
        :param obj_class_names: list of names ObjClass objects to delete
        :return: ProjectMeta class object
        '''
        res_items = self._delete_items(self._obj_classes, obj_class_names)
        return self.clone(obj_classes=ObjClassCollection(res_items))

    def delete_tag_meta(self, tag_name):
        '''
        The function delete_tag_meta delete tag with given name from ProjectMeta collection that stores TagMeta instances and return copy of ProjectMeta
        :param tag_name: str(name of TagMeta to detele from collection)
        :return: ProjectMeta class object
        '''
        return self.delete_tag_metas([tag_name])

    def delete_tag_metas(self, tag_names):
        '''
        The function delete_tag_metas delete tags with given list of names from ProjectMeta collection that stores TagMeta instances and return copy of ProjectMeta
        :param tag_names: list of names TagMeta objects to delete
        :return: ProjectMeta class object
        '''
        res_items = self._delete_items(self._tag_metas, tag_names)
        return self.clone(tag_metas=TagMetaCollection(res_items))

    def get_obj_class(self, obj_class_name):
        '''
        :param obj_class_name: str
        :return: ObjClass class object with given name from ProjectMeta collection that stores ObjClass instances
        '''
        return self._obj_classes.get(obj_class_name)

    def get_tag_meta(self, tag_name):
        '''
        :param tag_name: str
        :return: TagMeta class object with given name from ProjectMeta collection that stores TagMeta instances
        '''
        return self._tag_metas.get(tag_name)

    @staticmethod
    def merge_list(metas):
        '''
        The function merge_list merge metas from given list of metas in single ProjectMeta class object
        :param metas: list of ProjectMeta objects
        :return: ProjectMeta class object
        '''
        res_meta = ProjectMeta()
        for meta in metas:
            res_meta = res_meta.merge(meta)
        return res_meta

    def __str__(self):
        result = 'ProjectMeta:\n'
        result += 'Object Classes\n{}\n'.format(str(self._obj_classes))
        result += 'Tags\n{}\n'.format(str(self._tag_metas))
        return result


# coding: utf-8

import os
import requests
import json

from requests_toolbelt import MultipartEncoderMonitor, MultipartEncoder

import supervisely_lib.api.team_api as team_api
import supervisely_lib.api.workspace_api as workspace_api
import supervisely_lib.api.project_api as project_api
import supervisely_lib.api.neural_network_api as neural_network_api
import supervisely_lib.api.task_api as task_api
import supervisely_lib.api.dataset_api as dataset_api
import supervisely_lib.api.image_api as image_api
import supervisely_lib.api.annotation_api as annotation_api
import supervisely_lib.api.plugin_api as plugin_api
import supervisely_lib.api.agent_api as agent_api
import supervisely_lib.api.role_api as role_api
import supervisely_lib.api.user_api as user_api
import supervisely_lib.api.labeling_job_api as labeling_job_api
import supervisely_lib.api.video.video_api as video_api
import supervisely_lib.api.pointcloud.pointcloud_api as pointcloud_api
import supervisely_lib.api.object_class_api as object_class_api
import supervisely_lib.api.report_api as report_api
import supervisely_lib.api.app_api as app_api
import supervisely_lib.api.file_api as file_api
import supervisely_lib.api.image_annotation_tool_api as image_annotation_tool_api
import supervisely_lib.api.advanced_api as advanced_api
import supervisely_lib.api.import_storage_api as import_stoarge_api

from supervisely_lib.sly_logger import logger


from supervisely_lib.io.network_exceptions import process_requests_exception, process_unhandled_request

SUPERVISELY_TASK_ID = 'SUPERVISELY_TASK_ID'
SUPERVISELY_PUBLIC_API_RETRIES = 'SUPERVISELY_PUBLIC_API_RETRIES'
SUPERVISELY_PUBLIC_API_RETRY_SLEEP_SEC = 'SUPERVISELY_PUBLIC_API_RETRY_SLEEP_SEC'
SERVER_ADDRESS = 'SERVER_ADDRESS'
API_TOKEN = 'API_TOKEN'


class Api:
    def __init__(self, server_address, token, retry_count=None, retry_sleep_sec=None, external_logger=None, ignore_task_id=False):
        '''
        :param server_address: str (example: http://192.168.1.69:5555)
        :param token: str
        :param retry_count: int
        :param retry_sleep_sec: int
        :param external_logger: logger class object
        '''
        if token is None:
            raise ValueError("Token is None")
        self.server_address = server_address.strip('/')
        if ('http://' not in self.server_address) and ('https://' not in self.server_address):
            self.server_address = 'http://' + self.server_address

        if retry_count is None:
            retry_count = int(os.getenv(SUPERVISELY_PUBLIC_API_RETRIES, '20'))
        if retry_sleep_sec is None:
            retry_sleep_sec = int(os.getenv(SUPERVISELY_PUBLIC_API_RETRY_SLEEP_SEC, '1'))

        if len(token) != 128:
            raise ValueError("Invalid token {!r}: length != 128".format(token))

        self.headers = {'x-api-key': token}
        self.task_id = os.getenv(SUPERVISELY_TASK_ID)
        if self.task_id is not None and ignore_task_id is False:
            self.headers['x-task-id'] = self.task_id
        self.context = {}
        self.additional_fields = {}

        self.team = team_api.TeamApi(self)
        self.workspace = workspace_api.WorkspaceApi(self)
        self.project = project_api.ProjectApi(self)
        self.model = neural_network_api.NeuralNetworkApi(self)
        self.task = task_api.TaskApi(self)
        self.dataset = dataset_api.DatasetApi(self)
        self.image = image_api.ImageApi(self)
        self.annotation = annotation_api.AnnotationApi(self)
        self.plugin = plugin_api.PluginApi(self)
        self.agent = agent_api.AgentApi(self)
        self.role = role_api.RoleApi(self)
        self.user = user_api.UserApi(self)
        self.labeling_job = labeling_job_api.LabelingJobApi(self)
        self.video = video_api.VideoApi(self)
        # self.project_class = project_class_api.ProjectClassApi(self)
        self.object_class = object_class_api.ObjectClassApi(self)
        self.report = report_api.ReportApi(self)
        self.pointcloud = pointcloud_api.PointcloudApi(self)
        self.app = app_api.AppApi(self)
        self.file = file_api.FileApi(self)
        self.img_ann_tool = image_annotation_tool_api.ImageAnnotationToolApi(self)
        self.advanced = advanced_api.AdvancedApi(self)
        self.import_storage = import_stoarge_api.ImportStorageApi(self)

        self.retry_count = retry_count
        self.retry_sleep_sec = retry_sleep_sec

        self.logger = external_logger or logger

    @classmethod
    def from_env(cls, retry_count=5, ignore_task_id=False):
        '''
        :return: Api class object with server adress and token obtained from environment variables
        '''
        return cls(os.environ[SERVER_ADDRESS], os.environ[API_TOKEN], retry_count=retry_count, ignore_task_id=ignore_task_id)

    def add_header(self, key, value):
        '''
        Add given key and value to headers dictionary. Raise error if key is already set.
        :param key: str
        :param value: str
        '''
        if key in self.headers:
            raise RuntimeError(f'Header {key!r} is already set for the API object. '
                               f'Current value: {self.headers[key]!r}. Tried to set value: {value!r}')
        self.headers[key] = value

    def add_additional_field(self, key, value):
        '''
        Add given key and value to additional_fields dictionary.
        :param key: str
        :param value: str
        '''
        self.additional_fields[key] = value

    def post(self, method, data, retries=None, stream=False):
        '''
        Performs POST request to server with given parameters. Raise error if can not connect to server.
        :param method: str
        :param data: dict
        :param retries: int (number of attempts to access the server)
        :param stream: bool
        :return: Request class object
        '''
        if retries is None:
            retries = self.retry_count

        for retry_idx in range(retries):
            url = self.server_address + '/public/api/v3/' + method
            response = None
            try:
                if type(data) is bytes:
                    response = requests.post(url, data=data, headers=self.headers, stream=stream)
                elif type(data) is MultipartEncoderMonitor or type(data) is MultipartEncoder:
                    response = requests.post(url, data=data,
                                             headers={**self.headers, 'Content-Type': data.content_type}, stream=stream)
                else:
                    json_body = data
                    if type(data) is dict:
                        json_body = {**data, **self.additional_fields}
                    response = requests.post(url, json=json_body, headers=self.headers, stream=stream)

                if response.status_code != requests.codes.ok:
                    Api._raise_for_status(response)
                return response
            except requests.RequestException as exc:
                process_requests_exception(self.logger, exc, method, url,
                                           verbose=True, swallow_exc=True, sleep_sec=self.retry_sleep_sec,
                                           response=response,
                                           retry_info={"retry_idx": retry_idx + 1,
                                                       "retry_limit": retries})
            except Exception as exc:
                process_unhandled_request(self.logger, exc)
        raise requests.exceptions.RetryError("Retry limit exceeded ({!r})".format(url))

    def get(self, method, params, retries=None, stream=False, use_public_api=True):
        '''
        Performs GET request to server with given parameters. Raise error if can not connect to server.
        :param method: str
        :param params: dict
        :param retries: int (number of attempts to access the server)
        :param stream: bool
        :param use_public_api: bool
        :return: Request class object
        '''
        if retries is None:
            retries = self.retry_count

        for retry_idx in range(retries):
            url = self.server_address + '/public/api/v3/' + method
            if use_public_api is False:
                url = os.path.join(self.server_address, method)
            response = None
            try:
                json_body = params
                if type(params) is dict:
                    json_body = {**params, **self.additional_fields}
                response = requests.get(url, params=json_body, headers=self.headers, stream=stream)

                if response.status_code != requests.codes.ok:
                    Api._raise_for_status(response)
                return response
            except requests.RequestException as exc:
                process_requests_exception(self.logger, exc, method, url,
                                           verbose=True, swallow_exc=True, sleep_sec=self.retry_sleep_sec,
                                           response=response,
                                           retry_info={"retry_idx": retry_idx + 2,
                                                       "retry_limit": retries})
            except Exception as exc:
                process_unhandled_request(self.logger, exc)

    @staticmethod
    def _raise_for_status(response):
        '''
        Raise error and show message with code of mistake if given response can not connect to server.
        :param response: Request class object
        '''
        """Raises stored :class:`HTTPError`, if one occurred."""
        http_error_msg = ''
        if isinstance(response.reason, bytes):
            try:
                reason = response.reason.decode('utf-8')
            except UnicodeDecodeError:
                reason = response.reason.decode('iso-8859-1')
        else:
            reason = response.reason

        if 400 <= response.status_code < 500:
            http_error_msg = u'%s Client Error: %s for url: %s (%s)' % (response.status_code, reason, response.url, response.content.decode('utf-8'))

        elif 500 <= response.status_code < 600:
            http_error_msg = u'%s Server Error: %s for url: %s (%s)' % (response.status_code, reason, response.url, response.content.decode('utf-8'))

        if http_error_msg:
            raise requests.exceptions.HTTPError(http_error_msg, response=response)

    @staticmethod
    def parse_error(response, default_error="Error", default_message="please, contact administrator"):
        '''

        :param response: Request class object
        :param default_error: str
        :param default_message: str
        :return: number of error and message about curren connection mistake
        '''
        ERROR_FIELD = "error"
        MESSAGE_FIELD = "message"
        DETAILS_FIELD = "details"

        try:
            data_str = response.content.decode('utf-8')
            data = json.loads(data_str)
            error = data.get(ERROR_FIELD, default_error)
            details = data.get(DETAILS_FIELD, {})
            if type(details) is dict:
                message = details.get(MESSAGE_FIELD, default_message)
            else:
                message = details[0].get(MESSAGE_FIELD, default_message)

            return error, message
        except Exception as e:
            return "", ""



# coding: utf-8
import numpy as np

from supervisely_lib.geometry.constants import DATA, ORIGIN, GEOMETRY_SHAPE, GEOMETRY_TYPE, \
                                               LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID
from supervisely_lib.geometry.geometry import Geometry
from supervisely_lib.geometry.point_location import PointLocation
from supervisely_lib.geometry.rectangle import Rectangle
from supervisely_lib.imaging.image import resize_inter_nearest, restore_proportional_size


# TODO: rename to resize_bitmap_and_origin
def resize_origin_and_bitmap(origin: PointLocation, bitmap: np.ndarray, in_size, out_size):
    '''
    Change PointLocation and resize numpy array to match a certain size
    :return: new PointLocation class object and numpy array
    '''
    new_size = restore_proportional_size(in_size=in_size, out_size=out_size)

    row_scale = new_size[0] / in_size[0]
    col_scale = new_size[1] / in_size[1]

    # TODO: Double check (+restore_proportional_size) or not? bitmap.shape and in_size are equal?
    # Make sure the resulting size has at least one pixel in every direction (i.e. limit the shrinkage to avoid having
    # empty bitmaps as a result).
    scaled_rows = max(round(bitmap.shape[0] * row_scale), 1)
    scaled_cols = max(round(bitmap.shape[1] * col_scale), 1)

    scaled_origin = PointLocation(row=round(origin.row * row_scale), col=round(origin.col * col_scale))
    scaled_bitmap = resize_inter_nearest(bitmap, (scaled_rows, scaled_cols))
    return scaled_origin, scaled_bitmap


class BitmapBase(Geometry):
    def __init__(self, data: np.ndarray, origin: PointLocation = None, expected_data_dims=None,
                 sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        super().__init__(sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)
        """
        :param origin: PointLocation class object
        :param data: np.ndarray
        """
        if origin is None:
            origin = PointLocation(row=0, col=0)

        if not isinstance(origin, PointLocation):
            raise TypeError('BitmapBase "origin" argument must be "PointLocation" object!')

        if not isinstance(data, np.ndarray):
            raise TypeError('BitmapBase "data" argument must be numpy array object!')

        data_dims = len(data.shape)
        if expected_data_dims is not None and data_dims != expected_data_dims:
            raise ValueError('BitmapBase "data" argument must be a {}-dimensional numpy array. '
                             'Instead got {} dimensions'.format(expected_data_dims, data_dims))

        self._origin = origin.clone()
        self._data = data.copy()

    @classmethod
    def _impl_json_class_name(cls):
        """Descendants must implement this to return key string to look up serialized representation in a JSON dict."""
        raise NotImplementedError()

    @staticmethod
    def base64_2_data(s: str) -> np.ndarray:
        raise NotImplementedError()

    @staticmethod
    def data_2_base64(data: np.ndarray) -> str:
        raise NotImplementedError()

    def to_json(self):
        '''
        The function to_json convert bitmap to json format
        :return: Bitmap in json format(dict)
        '''
        res = {
            self._impl_json_class_name(): {
                ORIGIN: [self.origin.col, self.origin.row],
                DATA: self.data_2_base64(self.data)
            },
            GEOMETRY_SHAPE: self.geometry_name(),
            GEOMETRY_TYPE: self.geometry_name(),
        }
        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, json_data):
        '''
        The function from_json convert bitmap from json format to Bitmap class object.
        :param json_data: input bitmap in json format
        :return: Bitmap class object
        '''
        json_root_key = cls._impl_json_class_name()
        if json_root_key not in json_data:
            raise ValueError(
                'Data must contain {} field to create MultichannelBitmap object.'.format(json_root_key))

        if ORIGIN not in json_data[json_root_key] or DATA not in json_data[json_root_key]:
            raise ValueError('{} field must contain {} and {} fields to create MultichannelBitmap object.'.format(
                json_root_key, ORIGIN, DATA))

        col, row = json_data[json_root_key][ORIGIN]
        data = cls.base64_2_data(json_data[json_root_key][DATA])

        labeler_login = json_data.get(LABELER_LOGIN, None)
        updated_at = json_data.get(UPDATED_AT, None)
        created_at = json_data.get(CREATED_AT, None)
        sly_id = json_data.get(ID, None)
        class_id = json_data.get(CLASS_ID, None)
        return cls(data=data, origin=PointLocation(row=row, col=col),
                   sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    @property
    def origin(self) -> PointLocation:
        '''
        The function origin return copy of Bitmap class object PointLocation
        :return: PointLocation class object
        '''
        return self._origin.clone()

    @property
    def data(self) -> np.ndarray:
        '''
        The function data return copy of Bitmap class object data(numpy array)
        :return: numpy array
        '''
        return self._data.copy()

    def translate(self, drow, dcol):
        '''
        The function translate shifts the object by a certain number of pixels and return the copy of the current Bitmap object
        :param drow: horizontal shift
        :param dcol: vertical shift
        :return: Bitmap class object with new PointLocation
        '''
        translated_origin = self.origin.translate(drow, dcol)
        return self.__class__(data=self.data, origin=translated_origin)

    def fliplr(self, img_size):
        '''
        The function fliplr flip the current Bitmap object in horizontal and return the copy of the
        current Bitmap object
        :param img_size: size of the image
        :return: Bitmap class object with new data(numpy array) and PointLocation
        '''
        flipped_mask = np.flip(self.data, axis=1)
        flipped_origin = PointLocation(row=self.origin.row, col=(img_size[1] - flipped_mask.shape[1] - self.origin.col))
        return self.__class__(data=flipped_mask, origin=flipped_origin)

    def flipud(self, img_size):
        '''
        The function fliplr flip the current Bitmap object in vertical and return the copy of the
        current Bitmap object
        :param img_size: size of the image
        :return: Bitmap class object with new data(numpy array) and PointLocation
        '''
        flipped_mask = np.flip(self.data, axis=0)
        flipped_origin = PointLocation(row=(img_size[0] - flipped_mask.shape[0] - self.origin.row), col=self.origin.col)
        return self.__class__(data=flipped_mask, origin=flipped_origin)

    def scale(self, factor):
        '''
        The function scale change scale of the current Bitmap object with a given factor
        :param factor: float scale parameter
        :return: Bitmap class object with new data(numpy array) and PointLocation
        '''
        new_rows = round(self._data.shape[0] * factor)
        new_cols = round(self._data.shape[1] * factor)
        mask = self._resize_mask(self.data, new_rows, new_cols)
        origin = self.origin.scale(factor)
        return self.__class__(data=mask, origin=origin)

    @staticmethod
    def _resize_mask(mask, out_rows, out_cols):
        return resize_inter_nearest(mask.astype(np.uint8), (out_rows, out_cols)).astype(np.bool)

    def to_bbox(self):
        '''
        The function to_bbox create Rectangle class object from current Bitmap class object
        :return: Rectangle class object
        '''
        return Rectangle.from_array(self._data).translate(drow=self._origin.row, dcol=self._origin.col)



# coding: utf-8
import time
from supervisely_lib.collection.str_enum import StrEnum
from supervisely_lib.api.module_api import ApiField, ModuleApi, RemoveableModuleApi, ModuleWithStatus, \
                                           WaitingTimeExceeded


class LabelingJobApi(RemoveableModuleApi, ModuleWithStatus):
    class Status(StrEnum):
        PENDING = 'pending'
        IN_PROGRESS = "in_progress"
        ON_REVIEW = "on_review"
        COMPLETED = "completed"
        STOPPED = 'stopped'

    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.NAME,
                ApiField.README,
                ApiField.DESCRIPTION,

                ApiField.TEAM_ID,
                ApiField.WORKSPACE_ID,
                ApiField.WORKSPACE_NAME,
                ApiField.PROJECT_ID,
                ApiField.PROJECT_NAME,
                ApiField.DATASET_ID,
                ApiField.DATASET_NAME,

                ApiField.CREATED_BY_ID,
                ApiField.CREATED_BY_LOGIN,
                ApiField.ASSIGNED_TO_ID,
                ApiField.ASSIGNED_TO_LOGIN,
                ApiField.REVIEWER_ID,
                ApiField.REVIEWER_LOGIN,

                ApiField.CREATED_AT,
                ApiField.STARTED_AT,
                ApiField.FINISHED_AT,
                ApiField.STATUS,
                ApiField.DISABLED,
                ApiField.IMAGES_COUNT,
                ApiField.FINISHED_IMAGES_COUNT,
                ApiField.REJECTED_IMAGES_COUNT,
                ApiField.ACCEPTED_IMAGES_COUNT,
                ApiField.PROGRESS_IMAGES_COUNT,

                ApiField.CLASSES_TO_LABEL,
                ApiField.TAGS_TO_LABEL,
                ApiField.IMAGES_RANGE,
                ApiField.OBJECTS_LIMIT_PER_IMAGE,
                ApiField.TAGS_LIMIT_PER_IMAGE,

                ApiField.FILTER_IMAGES_BY_TAGS,
                ApiField.INCLUDE_IMAGES_WITH_TAGS,
                ApiField.EXCLUDE_IMAGES_WITH_TAGS,

                ApiField.ENTITIES,
                ]

    @staticmethod
    def info_tuple_name():
        return 'LabelingJobInfo'

    def __init__(self, api):
        ModuleApi.__init__(self, api)

    def _convert_json_info(self, info: dict, skip_missing=True):
        if info is None:
            return None
        else:
            field_values = []
            for field_name in self.info_sequence():
                if field_name in [ApiField.INCLUDE_IMAGES_WITH_TAGS, ApiField.EXCLUDE_IMAGES_WITH_TAGS]:
                    continue
                value = None
                if type(field_name) is str:
                    if skip_missing is True:
                        value = info.get(field_name, None)
                    else:
                        value = info[field_name]
                elif type(field_name) is tuple:
                    for sub_name in field_name[0]:
                        if value is None:
                            if skip_missing is True:
                                value = info.get(sub_name, None)
                            else:
                                value = info[sub_name]
                        else:
                            value = value[sub_name]
                else:
                    raise RuntimeError('Can not parse field {!r}'.format(field_name))

                if field_name == ApiField.FILTER_IMAGES_BY_TAGS:
                    field_values.append(value)
                    include_images_with_tags = []
                    exclude_images_with_tags = []
                    for fv in value:
                        key = ApiField.NAME
                        if key not in fv:
                            key = 'title'
                        if fv['positive'] is True:
                            include_images_with_tags.append(fv[key])
                        else:
                            exclude_images_with_tags.append(fv[key])
                    field_values.append(include_images_with_tags)
                    field_values.append(exclude_images_with_tags)
                    continue
                elif field_name == ApiField.CLASSES_TO_LABEL or field_name == ApiField.TAGS_TO_LABEL:
                    value = []
                    for fv in value:
                        key = ApiField.NAME
                        if ApiField.NAME not in fv:
                            key = 'title'
                        value.append(fv[key])
                elif field_name == ApiField.IMAGES_RANGE:
                    value = (value['start'], value['end'])

                field_values.append(value)
            return self.InfoType(*field_values)

    def create(self,
               name,
               dataset_id,
               user_ids,
               readme=None,
               description=None,
               classes_to_label=None,
               objects_limit_per_image=None,
               tags_to_label=None,
               tags_limit_per_image=None,
               include_images_with_tags=None,
               exclude_images_with_tags=None,
               images_range=None,
               reviewer_id=None):
        '''
        Create labeling job by given user in given dataset
        :param name: str
        :param dataset_id: int
        :param user_ids: list of integers
        :param readme: str
        :param description: str
        :param classes_to_label: list
        :param objects_limit_per_image: int
        :param tags_to_label: list
        :param tags_limit_per_image: int
        :param include_images_with_tags: bool
        :param exclude_images_with_tags: bool
        :param images_range: list
        :return: list
        '''
        if classes_to_label is None:
            classes_to_label = []
        if tags_to_label is None:
            tags_to_label = []

        filter_images_by_tags = []
        if include_images_with_tags is not None:
            for tag_name in include_images_with_tags:
                filter_images_by_tags.append({'name': tag_name, 'positive': True})

        if exclude_images_with_tags is not None:
            for tag_name in exclude_images_with_tags:
                filter_images_by_tags.append({'name': tag_name, 'positive': False})

        if objects_limit_per_image is None:
            objects_limit_per_image = 0

        if tags_limit_per_image is None:
            tags_limit_per_image = 0

        data = {ApiField.NAME: name,
                ApiField.DATASET_ID: dataset_id,
                ApiField.USER_IDS: user_ids,
                #ApiField.DESCRIPTION: description,
                ApiField.META: {
                     'classes': classes_to_label,
                     'projectTags': tags_to_label,
                     'imageTags': filter_images_by_tags,
                     'imageFiguresLimit': objects_limit_per_image,
                     'imageTagsLimit': tags_limit_per_image,}
                }

        if readme is not None:
            data[ApiField.README] = str(readme)

        if description is not None:
            data[ApiField.DESCRIPTION] = str(description)

        if images_range is not None:
            if len(images_range) != 2:
                raise RuntimeError('images_range has to contain 2 elements (start, end)')
            images_range = {'start': images_range[0], 'end': images_range[1]}
            data[ApiField.META]['range'] = images_range

        if reviewer_id is not None:
            data[ApiField.REVIEWER_ID] = reviewer_id

        response = self._api.post('jobs.add', data)
        # created_jobs_json = response.json()

        created_jobs = []
        for job in response.json():
            created_jobs.append(self.get_info_by_id(job[ApiField.ID]))
        return created_jobs

    def get_list(self, team_id, created_by_id=None, assigned_to_id=None, project_id=None, dataset_id=None,
                 show_disabled=False):
        '''
        Get all labeling that were created by given user and were assigned to given assigned user in given team
        :param team_id: int
        :param created_by_id: int
        :param assigned_to_id:
        :param project_id:
        :param dataset_id:
        :return: list
        '''
        filters = []
        if created_by_id is not None:
            filters.append({"field": ApiField.CREATED_BY_ID[0][0], "operator": "=", "value": created_by_id})
        if assigned_to_id is not None:
            filters.append({"field": ApiField.ASSIGNED_TO_ID[0][0], "operator": "=", "value": assigned_to_id})
        if project_id is not None:
            filters.append({"field": ApiField.PROJECT_ID, "operator": "=", "value": project_id})
        if dataset_id is not None:
            filters.append({"field": ApiField.DATASET_ID, "operator": "=", "value": dataset_id})
        return self.get_list_all_pages('jobs.list', {ApiField.TEAM_ID: team_id,
                                                     "showDisabled": show_disabled,
                                                     ApiField.FILTER: filters})

    def stop(self, id):
        '''
        :param id: int
        :return: stop labeling job  job will be unavailable for labeler with given id
        '''
        self._api.post('jobs.stop', {ApiField.ID: id})

    def get_info_by_id(self, id):
        '''
        :param id: int
        :return: labeling job with given id
        '''
        return self._get_info_by_id(id, 'jobs.info')

    def archive(self, id):
        '''
        Archive labeling job with given id
        :param id: int
        '''
        self._api.post('jobs.archive', {ApiField.ID: id})

    def get_status(self, id):
        '''
        :param id: int
        :return: status of labeling job with given id
        '''
        status_str = self.get_info_by_id(id).status
        return self.Status(status_str)

    def raise_for_status(self, status):
        #there is no ERROR status for labeling job
        pass

    def wait(self, id, target_status, wait_attempts=None, wait_attempt_timeout_sec=None):
        '''
        Awaiting achievement by given labeling job of a given status
        :param id: int
        :param target_status: status of labeling job we expect to destinate
        :param wait_attempts: int
        :param wait_attempt_timeout_sec: int (raise error if waiting time exceeded)
        :return: bool
        '''
        wait_attempts = wait_attempts or self.MAX_WAIT_ATTEMPTS
        effective_wait_timeout = wait_attempt_timeout_sec or self.WAIT_ATTEMPT_TIMEOUT_SEC
        for attempt in range(wait_attempts):
            status = self.get_status(id)
            self.raise_for_status(status)
            if status is target_status:
                return
            time.sleep(effective_wait_timeout)
        raise WaitingTimeExceeded('Waiting time exceeded')

    def get_stats(self, id):
        '''
        :param id: int
        :return: status of given labeling job
        '''
        response = self._api.post('jobs.stats', {ApiField.ID: id})
        return response.json()

    def get_activity(self, id):
        '''
        :param id: int
        :return: bool (True if labeling job is activity, False in overrise)
        '''
        job_info = self.get_info_by_id(id)
        df = self._api.project.get_activity(job_info.project_id)
        df = df[df.jobId == id]
        return df


# coding: utf-8

from supervisely_lib._utils import take_with_default
from supervisely_lib.video_annotation.key_id_map import KeyIdMap
from supervisely_lib.video_annotation.video_figure import VideoFigure
from supervisely_lib.pointcloud_annotation.pointcloud_object_collection import PointcloudObjectCollection


class PointcloudFigure(VideoFigure):
    """
    This is a class for creating and using PointcloudFigure objects for videos
    """
    def __init__(self, parent_object, geometry, frame_index=None, key=None,
                 class_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        :param parent_object:  VideoObject class object
        :param geometry: Geometry class object (cuboid_3d only)
        :param frame_index: int
        :param key: uuid class object
        '''
        super().__init__(parent_object, geometry, frame_index, key, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)
        #@TODO: validate geometry - allowed: only cuboid_3d + point_cloud

    def validate_bounds(self, img_size, _auto_correct=False):
        raise NotImplementedError()



# coding: utf-8

from collections import namedtuple
import os

from supervisely_lib.io.fs import file_exists, touch, dir_exists, list_files, get_file_name_with_ext
from supervisely_lib.imaging.image import SUPPORTED_IMG_EXTS
from supervisely_lib.io.json import dump_json_file, load_json_file
from supervisely_lib.project.project_meta import ProjectMeta
from supervisely_lib.task.progress import Progress
from supervisely_lib._utils import batched
from supervisely_lib.video_annotation.key_id_map import KeyIdMap

from supervisely_lib.api.module_api import ApiField
from supervisely_lib.collection.key_indexed_collection import KeyIndexedCollection
from supervisely_lib.video import video as sly_video

from supervisely_lib.project.project import Dataset, Project, OpenMode
from supervisely_lib.project.project import read_single_project as read_project_wrapper


from supervisely_lib.pointcloud_annotation.pointcloud_annotation import PointcloudAnnotation
from supervisely_lib.pointcloud import pointcloud as sly_pointcloud
from supervisely_lib.project.video_project import VideoDataset, VideoProject
from supervisely_lib.io.json import dump_json_file


PointcloudItemPaths = namedtuple('PointcloudItemPaths', ['pointcloud_path', 'related_images_dir', 'ann_path'])


class PointcloudDataset(VideoDataset):
    item_dir_name = 'pointcloud'
    related_images_dir_name = 'related_images'
    annotation_class = PointcloudAnnotation

    @staticmethod
    def _has_valid_ext(path: str) -> bool:
        return sly_pointcloud.has_valid_ext(path)

    def _get_empty_annotaion(self, item_name):
        return self.annotation_class()

    @staticmethod
    def _validate_added_item_or_die(item_path):
        # Make sure we actually received a valid image file, clean it up and fail if not so.
        try:
            sly_pointcloud.validate_format(item_path)
        except (sly_pointcloud.UnsupportedPointcloudFormat, sly_pointcloud.PointcloudReadException):
            os.remove(item_path)
            raise

    def get_related_images_path(self, item_name):
        item_name_temp = item_name.replace(".", "_")
        rimg_dir = os.path.join(self.directory, self.related_images_dir_name, item_name_temp)
        return rimg_dir

    def get_item_paths(self, item_name) -> PointcloudItemPaths:
        return PointcloudItemPaths(pointcloud_path=self.get_img_path(item_name),
                                   related_images_dir=self.get_related_images_path(item_name),
                                   ann_path=self.get_ann_path(item_name))

    def get_related_images(self, item_name):
        results = []
        path = self.get_related_images_path(item_name)
        if dir_exists(path):
            files = list_files(path, SUPPORTED_IMG_EXTS)
            for file in files:
                img_meta_path = os.path.join(path, get_file_name_with_ext(file)+".json")
                img_meta = {}
                if file_exists(img_meta_path):
                    img_meta = load_json_file(img_meta_path)
                results.append((file, img_meta))
        return results


class PointcloudProject(VideoProject):
    dataset_class = PointcloudDataset

    class DatasetDict(KeyIndexedCollection):
        item_type = PointcloudDataset

    @classmethod
    def read_single(cls, dir):
        return read_project_wrapper(dir, cls)


def download_pointcloud_project(api, project_id, dest_dir, dataset_ids=None, download_items=True, log_progress=False):
    LOG_BATCH_SIZE = 1

    key_id_map = KeyIdMap()

    project_fs = PointcloudProject(dest_dir, OpenMode.CREATE)

    meta = ProjectMeta.from_json(api.project.get_meta(project_id))
    project_fs.set_meta(meta)

    datasets_infos = []
    if dataset_ids is not None:
        for ds_id in dataset_ids:
            datasets_infos.append(api.dataset.get_info_by_id(ds_id))
    else:
        datasets_infos = api.dataset.get_list(project_id)

    for dataset in datasets_infos:
        dataset_fs = project_fs.create_dataset(dataset.name)
        pointclouds = api.pointcloud.get_list(dataset.id)

        ds_progress = None
        if log_progress:
            ds_progress = Progress('Downloading dataset: {!r}'.format(dataset.name), total_cnt=len(pointclouds))
        for batch in batched(pointclouds, batch_size=LOG_BATCH_SIZE):
            pointcloud_ids = [pointcloud_info.id for pointcloud_info in batch]
            pointcloud_names = [pointcloud_info.name for pointcloud_info in batch]

            ann_jsons = api.pointcloud.annotation.download_bulk(dataset.id, pointcloud_ids)

            for pointcloud_id, pointcloud_name, ann_json in zip(pointcloud_ids, pointcloud_names, ann_jsons):
                if pointcloud_name != ann_json[ApiField.NAME]:
                    raise RuntimeError("Error in api.video.annotation.download_batch: broken order")

                pointcloud_file_path = dataset_fs.generate_item_path(pointcloud_name)
                if download_items is True:
                    api.pointcloud.download_path(pointcloud_id, pointcloud_file_path)

                    related_images_path = dataset_fs.get_related_images_path(pointcloud_name)
                    related_images = api.pointcloud.get_list_related_images(pointcloud_id)
                    for rimage_info in related_images:
                        name = rimage_info[ApiField.NAME]
                        rimage_id = rimage_info[ApiField.ID]

                        path_img = os.path.join(related_images_path, name)
                        path_json = os.path.join(related_images_path, name + ".json")

                        api.pointcloud.download_related_image(rimage_id, path_img)
                        dump_json_file(rimage_info, path_json)

                else:
                    touch(pointcloud_file_path)

                dataset_fs.add_item_file(pointcloud_name,
                                         pointcloud_file_path,
                                         ann=PointcloudAnnotation.from_json(ann_json, project_fs.meta, key_id_map),
                                         _validate_item=False)

            ds_progress.iters_done_report(len(batch))

    project_fs.set_key_id_map(key_id_map)






# coding: utf-8

import cv2
import numpy as np

from shapely.geometry import mapping, LineString, Polygon as ShapelyPolygon
from supervisely_lib.geometry.conversions import shapely_figure_to_coords_list
from supervisely_lib.geometry.point_location import row_col_list_to_points
from supervisely_lib.geometry.vector_geometry import VectorGeometry
from supervisely_lib.geometry.constants import EXTERIOR, POINTS, LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID
from supervisely_lib.geometry import validation
from supervisely_lib import logger


class Polyline(VectorGeometry):
    '''
    This is a class for creating and using Polyline objects for Labels
    '''
    @staticmethod
    def geometry_name():
        return 'line'

    def __init__(self, exterior,
                 sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        """
        :param exterior: list of PointLocation objects
        """
        if len(exterior) < 2:
            raise ValueError('"{}" field must contain at least two points to create "Polyline" object.'
                             .format(EXTERIOR))

        super().__init__(exterior=exterior, interior=[], sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at,
                         created_at=created_at)

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert Polyline from json format to Poligon class object. If json format is not correct it generate exception error.
        :param data: input Polyline in json format
        :return: Polyline class object
        '''
        validation.validate_geometry_points_fields(data)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(exterior=row_col_list_to_points(data[POINTS][EXTERIOR], flip_row_col_order=True),
                   sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    def crop(self, rect):
        '''
        Crop the current Polyline with a given rectangle, if polyline cat't be cropped it generate exception error
        :param rect: Rectangle class object
        :return: list of Polyline class objects
        '''
        try:
            clipping_window = [[rect.top, rect.left], [rect.top, rect.right],
                               [rect.bottom, rect.right], [rect.bottom, rect.left]]
            clipping_window_shpl = ShapelyPolygon(clipping_window)

            exterior = self.exterior_np
            intersections_polygon = LineString(exterior).intersection(clipping_window_shpl)
            mapping_shpl = mapping(intersections_polygon)
        except Exception:
            logger.warn('Line cropping exception, shapely.', exc_info=False)
            raise

        res_lines_pts = shapely_figure_to_coords_list(mapping_shpl)

        # tiny hack to combine consecutive segments
        lines_combined = []
        for simple_l in res_lines_pts:
            if len(lines_combined) > 0:
                prev = lines_combined[-1]
                if prev[-1] == simple_l[0]:
                    lines_combined[-1] = list(prev) + list(simple_l[1:])
                    continue
            lines_combined.append(simple_l)

        return [Polyline(row_col_list_to_points(line)) for line in lines_combined]

    def _draw_impl(self, bitmap: np.ndarray, color, thickness=1, config=None):
        self._draw_contour_impl(bitmap, color, thickness, config=config)

    def _draw_contour_impl(self, bitmap: np.ndarray, color, thickness=1, config=None):
        exterior = self.exterior_np[:, ::-1]
        cv2.polylines(bitmap, pts=[exterior], isClosed=False, color=color, thickness=thickness)

    @property
    def area(self):
        return 0.0

    def approx_dp(self, epsilon):
        '''
        The function approx_dp approximates a polygonal curve with the specified precision
        :param epsilon: Parameter specifying the approximation accuracy. This is the maximum distance between the original curve and its approximation
        :return: Polyline class object
        '''
        exterior_np = self._approx_ring_dp(self.exterior_np, epsilon, closed=True).tolist()
        exterior = row_col_list_to_points(exterior_np, do_round=True)
        return Polyline(exterior)

    @classmethod
    def allowed_transforms(cls):
        from supervisely_lib.geometry.any_geometry import AnyGeometry
        from supervisely_lib.geometry.rectangle import Rectangle
        from supervisely_lib.geometry.bitmap import Bitmap
        from supervisely_lib.geometry.polygon import Polygon
        return [AnyGeometry, Rectangle, Bitmap, Polygon]


# coding: utf-8

from copy import deepcopy

from supervisely_lib.sly_logger import logger
from supervisely_lib.annotation.tag_meta import TagValueType
from supervisely_lib.metric.metric_base import MetricsBase
from supervisely_lib.metric.common import log_line, safe_ratio, sum_counters, TRUE_POSITIVE, TRUE_NEGATIVE, \
    FALSE_POSITIVE, FALSE_NEGATIVE, ACCURACY, PRECISION, RECALL, F1_MEASURE


RAW_COUNTERS = [TRUE_POSITIVE, TRUE_NEGATIVE, FALSE_POSITIVE, FALSE_NEGATIVE]


class ClassificationMetrics(MetricsBase):
    def __init__(self, tags_mapping, confidence_threshold=0):
        if len(tags_mapping) < 1:
            raise RuntimeError('At least one tags pair should be defined!')
        self._tags_mapping = tags_mapping.copy()
        self._confidence_threshold = confidence_threshold
        self._counters = {tag_name_gt: {counter: 0 for counter in RAW_COUNTERS} for tag_name_gt in
                          self._tags_mapping.keys()}

    def _classification_metrics(self, ann_1, ann_2):

        def is_passes_confidence_threshold(tag):
            if tag.meta.value_type == TagValueType.NONE:
                return True
            elif tag.meta.value_type == TagValueType.ANY_NUMBER:
                return tag.value >= self._confidence_threshold
            elif tag.meta.value_type == TagValueType.ANY_STRING or tag.meta.value_type == TagValueType.ONEOF_STRING:
                logger.warning("Classification tag '{}'".format(tag.name))
                return True

        current_metric_res = {}
        for tag_name_gt, tag_name_pred in self._tags_mapping.items():
            tag1 = ann_1.img_tags.get(tag_name_gt)
            tag2 = ann_2.img_tags.get(tag_name_pred)

            c1 = is_passes_confidence_threshold(tag1) if tag1 is not None else False
            c2 = is_passes_confidence_threshold(tag2) if tag2 is not None else False

            current_metric_res[tag_name_gt] = {
                TRUE_POSITIVE: int(c1 and c2),
                TRUE_NEGATIVE: int(not c1 and not c2),
                FALSE_POSITIVE: int(not c1 and c2),
                FALSE_NEGATIVE: int(c1 and not c2)
            }
        return current_metric_res

    def add_pair(self, ann_gt, ann_pred):
        res = self._classification_metrics(ann_gt, ann_pred)
        for tag_name_gt, met_data in res.items():
            for metric_name, metric_value in met_data.items():
                self._counters[tag_name_gt][metric_name] += metric_value

    @staticmethod
    def _calculate_complex_metrics(values):
        tp = values[TRUE_POSITIVE]
        tn = values[TRUE_NEGATIVE]
        fp = values[FALSE_POSITIVE]
        fn = values[FALSE_NEGATIVE]

        values[ACCURACY] = safe_ratio(tp + tn, tp + tn + fp + fn)
        values[PRECISION] = safe_ratio(tp, tp + fp)
        values[RECALL] = safe_ratio(tp, tp + fn)
        values[F1_MEASURE] = safe_ratio(2.0 * tp, 2.0 * tp + fp + fn)

    def get_metrics(self):
        result = deepcopy(self._counters)
        for pair_counters in result.values():
            self._calculate_complex_metrics(pair_counters)
        return result

    def get_total_metrics(self):
        result = sum_counters(self._counters.values(), (TRUE_POSITIVE, TRUE_NEGATIVE, FALSE_POSITIVE, FALSE_NEGATIVE))
        self._calculate_complex_metrics(result)
        return result

    def log_total_metrics(self):
        common_info = """
                P = condition positive (the number of real positive cases in the data)
                N = condition negative (the number of real negative cases in the data)
                TP = True Positive prediction
                TN = True Negative prediction
                FP = False Positive prediction (Type I error)
                FN = False Negative prediction (Type II error)
                Accuracy = (TP + TN)/(TP + TN + FP + FN) = TRUE/TOTAL
                Precision = TP / (TP + FP)
                Recall = TP / (TP + FN)
                F1-Measure = (2 * TP) / (2 * TP + FP + FN)
                """

        log_line()
        log_line(c='*')
        for line in common_info.split('\n'):
            line = line.strip()
            if len(line) > 0:
                logger.info(line.ljust(80))

        log_line(c='*')
        log_line()

        def print_evaluation_values(tag_pair_metrics):
            labels = [ACCURACY, PRECISION, RECALL, F1_MEASURE, TRUE_POSITIVE, TRUE_NEGATIVE, FALSE_POSITIVE,
                      FALSE_NEGATIVE]
            for label in labels:
                logger.info('    {0}:   {1:2.4f}'.format(label.ljust(16), tag_pair_metrics[label]))

        for i, (tag_name_gt, tag_metrics) in enumerate(self.get_metrics().items(), start=1):
            logger.info('{}) {} <--> {}:'.format(i, tag_name_gt, self._tags_mapping[tag_name_gt]))
            print_evaluation_values(tag_metrics)
            log_line()

        logger.info('Total values:')
        total_values = self.get_total_metrics()
        print_evaluation_values(total_values)
        log_line()

        log_line(c='*')




# coding: utf-8

import os
from supervisely_lib.api.module_api import ApiField
from supervisely_lib.api.task_api import TaskApi
from supervisely_lib._utils import take_with_default
from supervisely_lib.app.constants import DATA, STATE, CONTEXT, TEMPLATE
from supervisely_lib.io.fs import ensure_base_path
from supervisely_lib.task.progress import Progress
from supervisely_lib._utils import sizeof_fmt
from supervisely_lib import logger


class AppApi(TaskApi):
    def run_dtl(self, workspace_id, dtl_graph, agent_id=None):
        raise RuntimeError("Method is unavailable")

    def _run_plugin_task(self, task_type, agent_id, plugin_id, version, config, input_projects, input_models,
                         result_name):
        raise RuntimeError("Method is unavailable")

    def run_train(self, agent_id, input_project_id, input_model_id, result_nn_name, train_config=None):
        raise RuntimeError("Method is unavailable")

    def run_inference(self, agent_id, input_project_id, input_model_id, result_project_name, inference_config=None):
        raise RuntimeError("Method is unavailable")

    def get_training_metrics(self, task_id):
        raise RuntimeError("Method is unavailable")

    def deploy_model(self, agent_id, model_id):
        raise RuntimeError("Method is unavailable")

    def get_import_files_list(self, id):
        raise RuntimeError("Method is unavailable")

    def download_import_file(self, id, file_path, save_path):
        raise RuntimeError("Method is unavailable")

    def create_task_detached(self, workspace_id, task_type: str=None):
        raise RuntimeError("Method is unavailable")

    def upload_files(self, task_id, abs_paths, names, progress_cb=None):
        raise RuntimeError("Method is unavailable")

    def initialize(self, task_id, template, data=None, state=None):
        d = take_with_default(data, {})
        s = take_with_default(state, {})
        fields = [{"field": TEMPLATE, "payload": template}, {"field": DATA, "payload": d}, {"field": STATE, "payload": s}]
        resp = self._api.task.set_fields(task_id, fields)
        return resp

    def get_url(self, task_id):
        return os.path.join(self._api.server_address, "apps/sessions", str(task_id))

    def download_git_file(self, app_id, version, file_path, save_path):
        raise NotImplementedError()

    def download_git_archive(self, ecosystem_item_id, app_id, version, save_path, log_progress=True, ext_logger=None):
        payload = {
            ApiField.ECOSYSTEM_ITEM_ID: ecosystem_item_id,
            ApiField.VERSION: version,
            "isArchive": True
        }
        if app_id is not None:
            payload[ApiField.APP_ID] = app_id

        response = self._api.post('ecosystem.file.download', payload, stream=True)
        if log_progress:
            if ext_logger is None:
                ext_logger = logger

            length = -1
            # Content-Length
            if "Content-Length" in response.headers:
                length = int(response.headers['Content-Length'])
            progress = Progress("Downloading: ", length, ext_logger=ext_logger, is_size=True)

        mb1 = 1024 * 1024
        ensure_base_path(save_path)
        with open(save_path, 'wb') as fd:
            log_size = 0
            for chunk in response.iter_content(chunk_size=mb1):
                fd.write(chunk)
                log_size += len(chunk)
                if log_progress and log_size > mb1:
                    progress.iters_done_report(log_size)
                    log_size = 0

    def get_info(self, module_id):
        response = self._api.post('ecosystem.info', {ApiField.ID: module_id})
        return response.json()


# coding: utf-8
import os
from supervisely_lib.api.module_api import ModuleApiBase, ApiField
from supervisely_lib.io.fs import ensure_base_path, get_file_name_with_ext
from requests_toolbelt import MultipartEncoder
import mimetypes


class ImportStorageApi(ModuleApiBase):
    def get_meta_by_hashes(self, hashes):
        response = self._api.post('import-storage.internal.meta.list', {ApiField.HASHES: hashes})
        return response.json()



# coding: utf-8

from supervisely_lib.collection.str_enum import StrEnum


class WidgetType(StrEnum):
    TABLE = 'table'
    PLOTLY = "plotly"
    MARKDOWN = "markdown"
    NOTIFICATION = "notification"
    FORM = "form"
    BUTTON = "button"
    GALLERY = "gallery"
    LINECHART = "line-chart"


# coding: utf-8

from supervisely_lib.api.module_api import ApiField


class BaseWidget:
    widget_type = None

    @classmethod
    def _validate_type(cls, data):
        if cls.widget_type is None:
            raise ValueError("Child class has to define \"widget_type\" field")

        curr_type = data[ApiField.TYPE]
        if curr_type != str(cls.widget_type):
            raise ValueError("widget type is {!r}, but has to be {!r}".format(curr_type, str(cls.widget_type)))



# # coding: utf-8
# import os
# from supervisely_lib.file_cache.fs_caches import NNStorage, ImageStorage
#
#
# class FileCache:
#     def __init__(self, storage_root):
#         self._common_dir = constants.AGENT_STORAGE_DIR()
#
#         def create_st(stor_type, **kwargs):
#             if constants.WITH_LOCAL_STORAGE():
#                 sly.fs.mkdir(kwargs['storage_root'])
#                 return stor_type(**kwargs)
#             else:
#                 return EmptyStorage(**kwargs)
#
#         self._nn = create_st(NNStorage, name='NNs', storage_root=os.path.join(self._common_dir, 'neural_networks'))
#         self._image = create_st(ImageStorage, name='ImgCache', storage_root=os.path.join(self._common_dir, 'images'))
#
#     @property
#     def nn(self):
#         return self._nn
#
#     @property
#     def image(self):
#         return self._image


# coding: utf-8

from supervisely_lib.collection.str_enum import StrEnum


class ProjectType(StrEnum):
    IMAGES = 'images'
    VIDEOS = 'videos'
    VOLUMES = 'volumes'
    POINT_CLOUDS = 'point_clouds'


# coding: utf-8

from supervisely_lib.annotation.tag_collection import TagCollection
from supervisely_lib.video_annotation.video_tag import VideoTag


class VideoTagCollection(TagCollection):
    '''
    Collection that stores VideoTag instances.
    '''
    item_type = VideoTag

    def to_json(self, key_id_map=None):
        '''
        Converts collection to json serializable list.
        :param key_id_map: KeyIdMap class object
        :return: list of VideoTag in json format
        '''
        return [tag.to_json(key_id_map) for tag in self]

    @classmethod
    def from_json(cls, data, tag_meta_collection, key_id_map=None):
        '''
        Creates collection from json serializable list
        :param data: list of VideoTags in json format
        :param tag_meta_collection: TagMetaCollection
        :param key_id_map: KeyIdMap class object
        :return: VideoTagCollection class object
        '''
        tags = [cls.item_type.from_json(tag_json, tag_meta_collection, key_id_map) for tag_json in data]
        return cls(tags)